"""
NRV-Cellular Level simulations
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
from ..nmod.fascicles import *
from ..nmod.axons import *
from ..nmod.unmyelinated import *
from ..nmod.myelinated import *
from ..nmod.thin_myelinated import *
from ..fmod.extracellular import *
from ..fmod.electrodes import *
from ..fmod.materials import *
from ..fmod.stimulus import *
from ..backend.MCore import *
from ..backend.log_interface import rise_error, rise_warning, pass_info


###############################################################
#########################  Loaders  ###########################
###############################################################

def load_any_fascicle(data, extracel_context=False, intracel_context=False, rec_context=False):
    """
    generate any kind of fascicle from a dictionary or a json file

    Parameters
    ----------
    data    : str or dict
        json file path or dictionary containing fascicle information
    """
    synchronize_processes()
    if type(data) == str:
        fasc_dic = json_load(data)
    else: 
        fasc_dic = data
    fasc = fascicle()
    fasc.load(fasc_dic, extracel_context=extracel_context, intracel_context=intracel_context, rec_context=rec_context)
    if extracel_context and rec_context:
        return fasc , fasc.extra_stim, fasc.recorder
    elif extracel_context:
        return fasc , fasc.extra_stim
    elif rec_context:
        return fasc , fasc.recorder
    else:
        return fasc

def load_any_axon(data, extracel_context=False, intracel_context=False, rec_context=False):
    """
    generate any kind of axon from a dictionary or a json file

    Parameters
    ----------
    data    : str or dict
        json file path or dictionary containing axon information
    """
    if type(data) == str:
        ax_dic = json_load(data)
    else: 
        ax_dic = data

    if ax_dic["myelinated"] is True:
        if ax_dic["thin"]:
            ax = thin_myelinated(0,0,1,10)
        else:
            ax = myelinated(0,0,1,10)
    elif ax_dic["myelinated"] is False:
        ax = unmyelinated(0,0,1,10)
    else:
        ax = axon(0,0,1,10)

    ax.load(ax_dic, extracel_context=extracel_context, intracel_context=intracel_context, rec_context=rec_context)
    if extracel_context and rec_context:
        return ax , ax.extra_stim, ax.recorder
    elif extracel_context:
        return ax , ax.extra_stim
    elif rec_context:
        return ax , ax.recorder
    else:
        return ax



"""
NRV-Cellular Level postprocessing
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import faulthandler
import matplotlib.pyplot as plt
import os

from ...backend.file_handler import json_dump, json_load, is_iterable
from ...backend.log_interface import rise_error, rise_warning, pass_info

from ..cell.CL_postprocessing import *

# enable faulthandler to ease 'segmentation faults' debug
faulthandler.enable()


#############################
####### File handeler #######
#############################



def ls_axons_results(dir_path):
    '''
    return list of axons simulation result files

    Parameters
    ----------
    dir_path     : str
        path the results directory

    Returns
    -------
    files        :list of str
        list of axons result files
    '''
    list_files = [file for file in os.listdir(dir_path) if file[0:9]=='sim_axon_']
    return(list_files)



def rm_file(file_path, verbose=True):
    """
    Delete file

    Parameters
    ----------
    file_path : str
        path and name of the file to remove
    verbose     : str
        pass information when file is deleted 
    """
    os.remove(file_path)
    pass_info('folowing file removed :' + file_path, verbose=verbose)

def rm_sim_dir(dir_path, verbose=True):
    """
    Delete directory
    Warning: use with caution deleted files cannot be recovered

    Parameters
    ----------
    file_path : str
        path and name of the file to remove
    verbose     : str
        pass information when file is deleted 
    """
    if os.path.exists(dir_path):

        # checking whether the folder is empty or not
        if len(os.listdir(dir_path)) == 0:
            # removing the file using the os.remove() method
            os.rmdir(dir_path)
            pass_info('folowing folder removed :' + dir_path, verbose=verbose)
        else:
            # messaging saying folder not empty
            if os.path.exists(dir_path+"00_Fascicle_config.json",):
                rm_file(dir_path+"00_Fascicle_config.json", verbose)

            for file in ls_axons_results(dir_path):
                rm_file(dir_path+file, verbose)
            if len(os.listdir(dir_path)) == 0:
                os.rmdir(dir_path)
                pass_info('folowing folder removed :' + dir_path, verbose=verbose)
            else:
                pass_info( "Folder contains files or folders which cannot be deleted", verbose=verbose)
    else:
        # file not found message
        pass_info("Folder not found in the directory", verbose=verbose)



#############################
##### Result processing #####
#############################

def fascicular_state(dir_path, save=False, saving_file="facsicular_state.json", delete_files=False,
    verbose=True):
    """
    Return each axon caracteristics (blocked, Onset response, ...)

    Parameters
    ----------
    dir_path     : str
        path the results directory
    save        : bool
        if True save result in json file
    saving_file : str
        if save is True path and name of the saving file

    Returns
    -------
    facsicular_state       : dict

    """

    fascicular = load_simulation_from_json(dir_path+'00_Fascicle_config.json')
    facsicular_state = {'-1': fascicular}

    for file in ls_axons_results(dir_path):
        if 'extra_stim' not in facsicular_state['-1']:
            facsicular_state['-1']['extra_stim'] = extra_stim_properties(dir_path+file)

        axon = axon_state(dir_path + file)
        facsicular_state[axon['ID']]=axon

        if delete_files:
            rm_file(dir_path + file, verbose)
    if save:
        save_axon_results_as_json(facsicular_state, saving_file)


    return facsicular_state



#############################
## VISUALIZATION FUNCTIONS ##
#############################

def plot_fasc_state(facsicular_state, fig, axes, contour_color='k', myel_color='r', unmyel_color='b', num=False):
    """
    plot the fascicle in the Y-Z plane (transverse section)

    Parameters
    ----------
    fig     : matplotlib.figure
        figure to display the fascicle
    axes    : matplotlib.axes
        axes of the figure to display the fascicle
    contour_color   : str
        matplotlib color string applied to the contour. Black by default
    myel_color      : str
        matplotlib color string applied to the myelinated axons. Red by default
    unmyel_color    : str
        matplotlib color string applied to the myelinated axons. Blue by default
    num             : bool
        if True, the index of each axon is displayed on top of the circle
    """

    fasc = facsicular_state['-1']
    colors = []
    alpha = []
    N = 0
    while N in facsicular_state:
        if facsicular_state[N]['block_state']:
            if facsicular_state[N]['onset_state']:
                colors += ['green']
                alpha += [facsicular_state[N]['onset number']]
            else:
                colors += ['bleu']
                alpha += [0]
        elif facsicular_state[N]['block_state'] is None:
            colors += ['red']
            alpha += [facsicular_state[N]['onset number']]
        else:
            if facsicular_state[N]['onset_state']:
                colors += ['orange']
                alpha += [facsicular_state[N]['onset number']]
            else:
                colors += ['lightgray']
                alpha += [0]
        N +=1

    alpha = [1-(i/(1+max(alpha))) for i in alpha]

    ## plot contour
    axes.plot(fasc['y_vertices'], fasc['z_vertices'], linewidth=2, color=contour_color)
    ## plot axons
    circles = []
    for k in range(N):
        if fasc['axons_type'][k] == 1: # myelinated
            circles.append(plt.Circle((fasc['axons_y'][k], fasc['axons_z'][k]),\
                fasc['axons_diameter'][k]/2, color=colors[k], fill=True, alpha=alpha[k]))
        else:
            circles.append(plt.Circle((fasc['axons_y'][k], fasc['axons_z'][k]),\
                fasc['axons_diameter'][k]/2, color=colors[k], fill=True, alpha=alpha[k]))
    for circle in circles:
        axes.add_patch(circle)
    if num:
        for k in range(N):
            axes.text(fasc['axons_y'][k], fasc['axons_z'][k], str(k))
    ## plot electrode(s) if existings
    if 'extra_stim' in fasc:
        electrode = fasc['extra_stim']
        axes.scatter(electrode['y'], electrode['z'], color='red')



"""
NRV-Cellular Level discretization methods
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq
(c) ETIS - University Cergy-Pontoise - CNRS


todo:
 - Blocking vs dt
 - Blocking vs nseg
 - Onset_response vs dt (ou pas)
 - Onset_response vs dt (ou pas)
 - Velocity vs nseg 
 - Spike th vs nseg unmyelinated

"""
import numpy as np
from scipy.optimize import minimize

from ...backend.log_interface import rise_error, rise_warning, pass_info

calibrated_unmyelinated_models=['Rattay_Aberham','Sundt','Tigerholm']
calibrated_myelinated_models=['MRG','Gaines_motor','Gaines_sensory']
#calibrated_simulation = ['Velocity','Spike_threshold','Blocking_threshold','Onset_response']
calibrated_simulation = ['Velocity','Spike_threshold']


def _diff_nseg_spike_threshold_myelinated(nseg,tol):
    """
    Computes the norm to minimize for a specified nseg - Spike Threshold Simulation - Myelinated models

    Parameters
    ----------
    nseg : int
        number of segment 

    tol: 
        desired tolerance

    Returns
    -------
    float
        norm to minimize
    """
    nseg_log = np.log(nseg)
    tol_log = np.log(tol)
    poly_coeff = [ 0.08914346, -1.2569447,   3.12523366]
    poly = np.poly1d(poly_coeff)
    val = np.exp(poly(nseg_log))
    return ((val - tol)**2)



def _diff_dt_ratio_spike_threshold_myelinated(dt_ratio,tol):
    """
    Computes the norm to minimize for a specified dt - Spike Threshold Simulation - Myelinated models

    Parameters
    ----------
    dt_ratio : float
        dt ratio 

    tol: 
        desired tolerance

    Returns
    -------
    float
        norm to minimize
    """
    dt_log = np.log(dt_ratio)
    tol_log = np.log(tol)
    poly_coeff = [-0.05832062, 0.95287425, -5.45710455, 12.26637458, -7.04848512]
    poly = np.poly1d(poly_coeff)
    val = np.exp(poly(dt_log))
    return ((val - tol)**2)

def _diff_dt_ratio_spike_threshold_unmyelinated(dt_ratio,tol):
    """
    Computes the norm to minimize for a specified dt - Spike Threshold Simulation - Unmyelinated models

    Parameters
    ----------
    dt_ratio : float
        dt ratio 

    tol: 
        desired tolerance

    Returns
    -------
    float
        norm to minimize
    """
    dt_log = np.log(dt_ratio)
    tol_log = np.log(tol)
    poly_coeff =[ 0.01751393, -0.25108173,  1.23328312, -3.34147125,  6.57882033]
    poly = np.poly1d(poly_coeff)
    val = np.exp(poly(dt_log))
    return ((val - tol)**2)

def _diff_dt_velocity_myelinated(dt,tol):
    """
    Computes the norm to minimize for a specified dt - Velocity estimation - Myelinated models

    Parameters
    ----------
    dt : float
        time discretization in ms

    tol: 
        desired tolerance

    Returns
    -------
    float
        norm to minimize
    """
    poly_coeff = [ 3.65238002e+05, -4.13999358e+04, 2.16625888e+03,-2.75739556e-01]
    poly = np.poly1d(poly_coeff)
    return ((poly(dt) - tol )**2)

def _diff_dt_velocity_sundt_tigerholm(dt,tol):
    """
    Computes the norm to minimize for a specified dt - Velocity estimation - Sundt-Tigerholm unmyelinated models

    Parameters
    ----------
    dt : float
        time discretization in ms

    tol: 
        desired tolerance

    Returns
    -------
    float
        norm to minimize
    """

    poly_coeff = [ 1.70847637e+05, -1.40276023e+04,  5.26842025e+02,  1.27772606e-01]
    poly = np.poly1d(poly_coeff)
    return ((poly(dt) - tol )**2)

def _diff_dt_velocity_rattay_aberham(dt,tol):
    """
    Computes the norm to minimize for a specified dt - Velocity estimation - Rattay-Aberham unmyelinated models

    Parameters
    ----------
    dt : float
        time discretization in ms

    tol: 
        desired tolerance

    Returns
    -------
    float
        norm to minimize
    """

    poly_coeff = [ 2.79799924e+05, -1.18439668e+04,  7.14708491e+02, -1.43308070e-01]
    poly = np.poly1d(poly_coeff)
    return ((poly(dt) - tol )**2)


def _get_dt_velocity_myelinated(tol,model):
    """
    returns the dt value for requested tolerance - myelinated models - velocity estimation

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model
    Returns
    -------
    dt
        time discretization in ms
    """
    if (tol > 50):
        rise_warning("Requested tolerance is out of calibration range. Returned value might be inacurate.")
    res = minimize(_diff_dt_velocity_myelinated, 1.0, args=(tol), method='Nelder-Mead', tol=1e-6)
    return(res.x[0])

def _get_dt_velocity_unmyelinated(tol,model):
    """
    returns the dt value for requested tolerance - unmyelinated models - velocity estimation

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model
    Returns
    -------
    dt
        time discretization in ms
    """
    if (model == 'Rattay_Aberham'):
        if (tol > 40):
            rise_warning("Requested tolerance is out of calibration range. Returned value might be inacurate.")
        res = minimize(_diff_dt_velocity_rattay_aberham, 1.0, args=(tol), method='Nelder-Mead', tol=1e-6)
    if (model == 'Sundt') or (model == 'Tigerholm'):
        if (tol > 10):
            rise_warning("Requested tolerance is out of calibration range. Returned value might be inacurate.")
        res = minimize(_diff_dt_velocity_sundt_tigerholm, 1.0, args=(tol), method='Nelder-Mead', tol=1e-6)
    return(res.x[0])


def _get_dt_spike_threshold_myelinated(tol,model):
    """
    returns the dt value for requested tolerance - myelinated models - Spike threshold

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model
    Returns
    -------
    dt_ratio
        time discretization ratio
    """
    if (tol > 10):
        rise_warning("Requested tolerance is out of calibration range. Returned value might be inacurate.")
    res = minimize(_diff_dt_ratio_spike_threshold_myelinated, 10.0, args=(tol), method='Nelder-Mead', tol=1e-6)
    return(res.x[0])


def _get_dt_spike_threshold_unmyelinated(tol,model):
    """
    returns the dt value for requested tolerance - unmyelinated models - Spike threshold

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model
    Returns
    -------
    dt_ratio
        time discretization ratio
    """
    if (tol > 40):
        rise_warning("Requested tolerance is out of calibration range. Returned value might be inacurate.")
    res = minimize(_diff_dt_ratio_spike_threshold_unmyelinated, 1.0, args=(tol), method='Nelder-Mead', tol=1e-6)
    return(res.x[0])

def _get_nseg_spike_threshold_myelinated(tol,model):
    """
    returns the number of segment for the requested tolerance - myelinated models - Spike threshold

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model
    Returns
    -------
    nseg: int
        number of segment
    """
    if (tol > 25):
        rise_warning("Requested tolerance is out of calibration range. Returned value might be inacurate.")
    res = minimize(_diff_nseg_spike_threshold_myelinated, 1.0, args=(tol), method='Nelder-Mead', tol=1e-6)
    return(res.x[0])

def _get_dt_velocity(tol,model):
    """
    returns the dt value for requested tolerance - Velocity estimation

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model
    Returns
    -------
    dt
        time discretization in ms
    """

    if (model in calibrated_myelinated_models):
        return(_get_dt_velocity_myelinated(tol,model))
    if (model in calibrated_unmyelinated_models):
        return(_get_dt_velocity_unmyelinated(tol,model))
    if (model not in (calibrated_myelinated_models) and model not in (calibrated_unmyelinated_models)):
        rise_error('Selected model is not calibrated yet.')
        return(0)

def _get_dt_spike_threshold(tol,model):
    """
    returns the dt value for requested tolerance - Spike threshold

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model
    Returns
    -------
    dt
        time discretization ratio
    """

    if (model in calibrated_myelinated_models):
        return(_get_dt_spike_threshold_myelinated(tol,model))
    if (model in calibrated_unmyelinated_models):
        return(_get_dt_spike_threshold_unmyelinated(tol,model))
    if (model not in (calibrated_myelinated_models) and model not in (calibrated_unmyelinated_models)):
        rise_error('Selected model is not calibrated yet.')

def _get_nseg_spike_threshold(tol,model,L=None,d=None):
    """
    returns the number of segment for requested tolerance - Spike threshold

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model

    d: float
        axon diameter in um (only for unmyelinated models)

    L: float
        Axon length in um (only for unmyelinated models)
    Returns
    -------
    nseg
        number of segment
    """

    if (model in calibrated_myelinated_models):
        return(_get_nseg_spike_threshold_myelinated(tol,model))
    if (model in calibrated_unmyelinated_models):
        rise_error('Not implemented yet.')
    if (model not in (calibrated_myelinated_models) and model not in (calibrated_unmyelinated_models)):
        rise_error('Selected model is not calibrated yet.')

def Get_nseg(tol,model,simulation_category,d=None,L=None):
    """
    Returns the number of segment for the requested tolerance

    Parameters
    ----------
    tol : float
        desired tolerance

    model: str
        simulated 

    simulation_category: str
        Simulation to perform 

    d: float
        axon diameter in um (only for unmyelinated models)

    L: float
        axon length in um (only for unmyelinated models)

    Returns
    -------
    nseg number of segments
    """

    if (simulation_category not in (calibrated_simulation)):
        rise_error('Selected simulation is not calibrated yet.')
    else:
        if (simulation_category == 'Velocity'):
            rise_error('Not implemented yet.')
        
        if (simulation_category == 'Spike_threshold'):
            nseg = _get_nseg_spike_threshold(tol,model)
            return(np.int32(np.round(nseg)))


def Get_dt(tol,model,simulation_category,stim_pw = None):
    """
    Estimates the dt value for requested tolerance

    Parameters
    ----------
    tol : float
        desired tolerance

    model: str
        simulated 

    simulation_category: str
        Simulation to perform 

    stim_pw: float
        Stimulation characteristic pulse width in ms (not required for velocity simulation)

    Returns
    -------
    dt
        time discretization value
    """

    if (simulation_category not in (calibrated_simulation)):
        rise_error('Selected simulation is not calibrated yet.')
    else:
        if (simulation_category == 'Velocity'):
            return(_get_dt_velocity(tol,model))

        if (simulation_category == 'Spike_threshold'):
            if (stim_pw>250e-3):
                rise_warning("Characteristic pulse width is out of calibration range. Returned value might be inacurate.")
            dt_ratio = stim_pw/_get_dt_spike_threshold(tol,model)
            return(dt_ratio)






from .log_interface import rise_error, rise_warning, pass_info

def set_attributes(my_object, attributes_dict):
    for key, value in attributes_dict.items():
        if key in my_object.__dict__:
            setattr(my_object, key, value)
        else:
            rise_warning('trying to set a non existing attribute '+str(key)+' in '+str(type(my_object)))
    return 0



"""
NRV-electrodes
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import faulthandler
import numpy as np
from ..backend.log_interface import rise_error, rise_warning, pass_info
from ..backend.file_handler import *
from ..backend.NRV_Class import NRV_class

# enable faulthandler to ease 'segmentation faults' debug
faulthandler.enable()

def is_FEM_electrode(elec):
    """
    Check if the electrode is a FEM based electrode
    """
    return issubclass(type(elec), FEM_electrode)


def is_CUFF_electrode(elec):
    """
    Check if the electrode is a LIFE electrode
    """
    return isinstance(elec, CUFF_electrode)

def is_LIFE_electrode(elec):
    """
    Check if the electrode is a LIFE electrode
    """
    return isinstance(elec, LIFE_electrode)

def is_analytical_electrode(elec):
    """
    Check if the electrode is an analytical based electrode
    """
    return not issubclass(type(elec), FEM_electrode)

def load_any_electrode(data):
    """
    return any kind of electrod context properties from a dictionary or a json file

    Parameters
    ----------
    data    : str or dict
        json file path or dictionary containing extracel_context information
    """
    if type(data) == str:
        elec_dic = json_load(data)
    else: 
        elec_dic = data

    if elec_dic["type"] == "electrode":
        elec = electrode()
    elif elec_dic["type"] == "point source":
        elec = point_source_electrode(0,0,0)
    elif elec_dic["type"] == "FEM":
        elec = FEM_electrode("")
    elif elec_dic["type"] == "LIFE":
        elec = LIFE_electrode("",0,0,0,0,0)
    elif elec_dic["type"] == "CUFF":
        elec = CUFF_electrode()
    elif elec_dic["type"] == "CUFF MP":
        elec = CUFF_MP_electrode()
    else:
        rise_error("Electrode type not recognizede")

    elec.load(elec_dic)
    return elec


class electrode(NRV_class):
    """
    Objet for generic electrode description. Each electrode has an ID and a position.

    """
    def __init__(self, ID=0):
        """
        Instantiation of a generic electrode

        Parameters
        ----------
        ID  : int
            electrode identification number, set to 0 by default
        """
        super().__init__()
        self.ID = ID
        self.footprint = np.asarray([])
        self.type = "electrode"
        self.is_multipolar = False

    def save_electrode(self, save=False, fname='electrode.json'):
        rise_warning('save_electrode is a deprecated method use save')
        self.save(save=save, fname=fname)
    def load_electrode(self, data='electrode.json'):
        rise_warning('load_electrode is a deprecated method use load')
        self.load(data=data)

    def get_ID_number(self):
        """
        get the ID of a electrode

        Returns
        -------
        ID : int
            identification number of the electrode
        """
        return self.ID

    def set_ID_number(self, ID):
        """
        set the identification number of an electrode

        Parameters
        ----------
        ID  : int
            electrode identification number
        """
        self.ID = ID

    def get_footptint(self):
        """
        get the footprint of a electrode

        Returns
        -------
        footprint : np.array
            identification number of the electrode
        """
        return self.footprint

    def set_footprint(self, footprint):
        """
        set the footprint of a electrode

        Parameters
        ----------
        footprint : np.array
            array contaning the electrod linear footprint (electrode response at 1mA)
            on several space points
        """
        self.footprint = np.array(footprint)

    def compute_field(self, I):
        """ Compute the external field using the Point source approximation

        Parameters
        ----------
        I       : float
            current value, in uA

        Returns
        -------
        v_ext   : np.array
            external voltage field value at specified coordinates with the specified material in mV
        """
        # convert uA in mA
        I_mA = I*1e-3
        v_ext = I_mA*self.footprint
        return v_ext

class point_source_electrode(electrode):
    """
    Point source electrode. Inherite from electrode. The electrode is punctual and act as a\
    monopole.
    """
    def __init__(self, x=0, y=0, z=0, ID=0):
        """
        Instantiation of a Point source electrode

        Parameters
        ----------
        x   : float
            x position of the electrode, in um
        y   : float
            y position of the electrode, in um
        z   : float
            z position of the electrode, in um
        ID  : int
            electrode identification number, set to 0 by default
        """
        super().__init__(ID)
        self.x = x
        self.y = y
        self.z = z

        self.type = "point source"

    def compute_footprint(self, x, y, z, mat):
        """
        Compute the linear footprint (electrode response at 1mA)

        Parameters
        ----------
        x       : float
            array like, x corrdinates where to compute the field, in um
        y       : float
            array like, y coordinates where to compute the field, in um
        z       : float
            array like, y coordinates where to compute the field, in um
        mat     : float
            material object, see Material help for more detail
        """
        if mat.is_isotropic():
            # 1e-6  on distances to stay in m (condivtivity specified in S/m
            self.footprint = 1./(4*np.pi*np.sqrt((1e-6*(self.x-x))**2 + (1e-6*(self.y-y))**2 +\
                (1e-6*(self.z-z))**2)*mat.sigma)
        else:
            sx = mat.sigma_yy * mat.sigma_zz
            sy = mat.sigma_xx * mat.sigma_zz
            sz = mat.sigma_xx * mat.sigma_yy
            # 1e-6  on distances to stay in m (condivtivity specified in S/m)
            self.footprint = 1./(4*np.pi*np.sqrt(sx*(1e-6*(self.x-x))**2 + \
                sy*(1e-6*(self.y-y))**2 + sz*(1e-6*(self.z-z))**2))

class FEM_electrode(electrode):
    """
    Electrode located in Finite Element Model in Comsol
    """
    def __init__(self, label, ID=0):
        """
        Instrantiation of a FEM electrode
        """
        super().__init__(ID)
        self.label = label
        self.ID = ID
        self.footprint = np.asarray([])
        self.is_volume = False
        self.type = "FEM"

    def set_footprint(self, V_1mA):
        """
        set the footprin of a FEM electrode

        Parameters:
        -----------
        V_1mA : list, array, numpy array
            Voltage response at 1mA
        """
        self.footprint = np.asarray(V_1mA)

class LIFE_electrode(FEM_electrode):
    """
    Longitudinal IntraFascicular Electrode for FEM models
    """
    def __init__(self, label="LIFE_1", D=25, length=1000, x_shift=0, y_c=0, z_c=0, ID=0):
        """
        Instantiation of a LIFE electrode

        Parameters
        ----------
        label   : str
            name of the electrode in the COMSOL file
        D       : float
            diameter of the electrode, in um
        length  : float
            length of the electrode, in um
        x_shift : float
            geometrical offset from the start (x=0) of the simulation
        y_c     : float
            y-coordinate of the center of the electrode, in um
        z_c     : float
            z-coordinate of the center of the electrode, in um
        """
        super().__init__(label, ID)
        self.D = D
        self.length = length
        self.x_shift = x_shift
        self.y_c = y_c
        self.z_c = z_c
        self.type = "LIFE"

    
    def parameter_model(self, model, res="default"):
        """
        Parameter the model electrode with user specified dimensions

        Parameters
        ----------
        model : obj
            FEM COMSOL or Fenics simulation to parameter, se FEM or Extracellular for more details
        """
        if model.type == 'COMSOL':
            model.set_parameter(self.label+'_D', str(self.D)+'[um]')
            model.set_parameter(self.label+'_Length', str(self.length)+'[um]')
            model.set_parameter(self.label+'_y_c', str(self.y_c)+'[um]')
            model.set_parameter(self.label+'_z_c', str(self.z_c)+'[um]')
            model.set_parameter(self.label+'_x_offset', str(self.x_shift)+'[um]')
        else:
            model.add_electrode(elec_type=self.type, x_c=self.x_shift+(self.length/2),\
            y_c=self.y_c, z_c=self.z_c, length=self.length, D=self.D, is_volume=self.is_volume,\
            res=res)



class CUFF_electrode(FEM_electrode):
    """
    CUFF electrode for FEM models
    """
    def __init__(self, label="", contact_length=100, is_volume=False, contact_thickness=None,\
        insulator=True, insulator_length=None, insulator_thickness=None, x_center=0,\
        insulator_offset=0, ID=0):
        """
        Instantiation of a LIFE electrode

        Parameters
        ----------
        label               : str
            name of the electrode in the COMSOL file
        x_center            :float
            x-position of the CUFF center in um, by default 0
            length of the CUFF electrod in um, by default 100
        contact_length      :float
            length along x of the contact site in um, by default 100
        is_volume   : bool 
            if True the contact is kept on the mesh as a volume, by default True
        contact_thickness   :float
            thickness of the contact site in um, by default 5
        insulator            :bool
            remove insulator ring from the mesh (no conductivity), by default True
        insulator_thickness :float
            thickness of the insulator ring in um, by default 20
        insulator_length    :float
            length along x of the insulator ring in um, by default 1000
        """
        super().__init__(label, ID)
        self.contact_length = contact_length
        self.contact_thickness = contact_thickness
        self.x_center = x_center
        self.is_volume = is_volume
        self.insulator = insulator
        self.insulator_length = insulator_length
        self.insulator_thickness = insulator_thickness
        self.insulator_offset = insulator_offset
        self.type = "CUFF"

    def parameter_model(self, model, res="default"):
        """
        Parameter the model electrode with user specified dimensions

        Parameters
        ----------
        model : obj
            FEM COMSOL or Fenics simulation to parameter, se FEM or Extracellular for more details
        """
        if model.type == 'COMSOL':
            model.set_parameter(self.label+'_contact_length', str(self.contact_length)+'[um]')
            model.set_parameter(self.label+'_x_center', str(self.x_center)+'[um]')
            if self.contact_thickness is not None:
                model.set_parameter(self.label+'_contact_thickness', str(self.contact_thickness)+'[um]')
            if self.insulator_thickness is not None:
                model.set_parameter(self.label+'_insulator_thickness', str(self.insulator_thickness)+'[um]')
            if self.insulator_length is not None:
                model.set_parameter(self.label+'_insulator_length', str(self.insulator_length)+'[um]')
        else:
            model.add_electrode(elec_type=self.type, insulator_length=self.insulator_length, is_volume=self.is_volume,\
                insulator=self.insulator, insulator_thickness=self.insulator_thickness, contact_length=self.contact_length,\
                contact_thickness=self.contact_thickness, insulator_offset=self.insulator_offset, x_c=self.x_center, res=res)

class CUFF_MP_electrode(CUFF_electrode):
    """
    MultiPolar CUFF electrode for FEM models
    """
    def __init__(self, label="CUFF_MP", N_contact=4, contact_width=None, contact_length=100,\
        is_volume=False, contact_thickness=None, insulator=True, insulator_length=None,\
        insulator_thickness=None, x_center=0, insulator_offset=0, ID=0):
        """
        Instantiation of a LIFE electrode

        Parameters
        ----------
        label               : str
            name of the electrode in the COMSOL file
        x_center            :float
            x-position of the CUFF center in um, by default 0
        N_contact           :int
            Number of contact site of the electrode, by default 4
            length of the CUFF electrod in um, by default 100
        contact_length      :float
            length along x of the contact site in um, by default 100
        is_volume   : bool 
            if True the contact is kept on the mesh as a volume, by default True
        contact_thickness   :float
            thickness of the contact site in um, by default 5
        inactive            :bool
            remove insulator ring from the mesh (no conductivity), by default True
        insulator_thickness :float
            thickness of the insulator ring in um, by default 20
        insulator_length    :float
            length along x of the insulator ring in um, by default 1000
        """
        super().__init__(label=label,contact_length=contact_length,is_volume=is_volume,\
            contact_thickness=contact_thickness, insulator_length=insulator_length,\
            insulator=insulator, insulator_thickness=insulator_thickness, x_center=x_center,\
            insulator_offset=insulator_offset, ID=ID)
        self.N_contact = N_contact
        self.contact_width = contact_width
        self.type = "CUFF MP"
        self.is_multipolar = True

    def parameter_model(self, model, res="default"):
        """
        Parameter the model electrode with user specified dimensions

        Parameters
        ----------
        model : obj
            FEM COMSOL or Fenics simulation to parameter, se FEM or Extracellular for more details
        """
        if model.type == 'COMSOL':
            rise_warning('Multipolar CUFF not implemented on comsol')
        else:
            model.add_electrode(elec_type=self.type, N=self.N_contact, is_volume=self.is_volume,\
                contact_width=self.contact_width, insulator_length=self.insulator_length,\
                insulator_thickness=self.insulator_thickness, contact_length=self.contact_length,\
                insulator=self.insulator, contact_thickness=self.contact_thickness,\
                x_c=self.x_center, insulator_offset=self.insulator_offset, res=res)



"""
NRV-compileMods
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import os
import inspect
from .log_interface import rise_warning

# test if neuron correctly installed
# warning: difficult as neuron can be 'mocked' on github and rtd for delploy and help
try:
    import neuron as nrn                # this still work with fake neuron (mock)
    nrn_version = nrn.__version__       # this fails :)
except:
    # no Neuron or fake neuron
    rise_warning('No mods compiled as the environnement variable NRVPATH does not exist')
else:
    # done anyway
    dir_path = os.environ['NRVPATH'] + '/_misc/mods/'
    ls = os.listdir(dir_path)
    test_mods = False
    i=0
    while (not test_mods) and i<len(ls):
        test_mods = not '.' in ls[i]
        i += 1

    def NeuronCompile():
        #path2compiled_mods =  dir_path + '/mods/x86_64'
        path2_mods = dir_path
        os.system('chmod +x '+ os.environ['NRVPATH'] +'/nrv2calm')
        os.system('cd ' +path2_mods+  '&& nrnivmodl')

    if not test_mods:
        print('Mods files are not compiled, executing nrnivmodl...')
        NeuronCompile()
        print('Compilation done')



"""
NRV-thin_myelinated
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import math
import numpy as np
from .axons import *
from ..backend.log_interface import rise_error, rise_warning, pass_info

def get_Adelta_parameters(diameter):
    """
    Compute the A delta parameters, see on code for exact scientific references.

    Attributes
    ----------
    diameter    : float
        diameter of the axon (or fiber) to create um. Note that for mathematical reason, this should be > 2

    Returns
    -------
    axonD   : float
        internal axon diameter under the myeline, in um
    nodeD   : float
        length of the nodes of Ranvier, in um
    paraD   : float
        length of the paranodes, in um
    deltax  : float
        internode length, in um
    nl      : float
        number of myelin sheaths in average
    """
    # taken from Berthold, Nilsson and Rydmark, to get the number of layer of myelin layers from axonD
    # from 6 months dorsal root cat
    # Berthold, C. H., Nilsson, I., & Rydmark, M. (1983). Axon diameter and myelin sheath thickness in nerve fibres of the
    # ventral spinal root of the seventh lumbar nerve of the adult and developing cat. Journal of anatomy, 136(Pt 3), 483.
    C0 = 24.72
    C1 = -5.73
    C2 = 155.44
    # thickness of a myelin layer, in um , from Berthold, Nilsson and Rydmark 1983
    k = 0.018
    # node diameter and corresponding number of myelin layer taken from Rydmark and Berthold
    # Berthold, C. H., & Rydmark, M. (1983). Electron microscopic serial section analysis of nodes of Ranvier in lumbosacral
    # spinal roots of the cat: ultrastructural organization of nodal compartments in fibres of different sizes. Journal of Neurocytology, 12(3), 475-505.
    nodeD_Rydmark = [0.8, 1.5, 2.6, 3.3, 4.3]
    nl_Rydmark = [27, 46, 88, 105, 123]
    nodeD_vs_nl_poly = np.poly1d(np.polyfit(nl_Rydmark, nodeD_Rydmark, 3))

    # step 1: get axonD
    def solver(D, C0, C1, C2, k):
        def fiber_axon_diameter_equation(axonD):
            return (2*k*C0 - D) + axonD*(1+2*k*C1) + 2*k*C2*np.log10(axonD)
        return fiber_axon_diameter_equation

    root = optimize.fsolve(solver(diameter, C0, C1, C2, k), 1)
    axonD = root[0]
    # step 2: get nl
    nl = C0 + C1*axonD + C2*np.log10(axonD)
    # step 3: get nodeD
    nodeD = nodeD_vs_nl_poly(nl)
    # step 4: get deltax
    # interpolation from Jacobs, J. M., & Love, S. (1985). Qualitative and quantitative morphology of human sural nerve at different ages. Brain, 108(4), 897-924.
    deltax = 80*(diameter-1)
    # step 5: paraD
    # arbitrary conical structure not possible in neuron, therefore, an average of axonD and nodeD to ensure transition
    # matter of choice, could also be changed to paraD = nodeD as in MRG model
    paraD = (nodeD + axonD)/2

    return float(axonD), float(nodeD), float(paraD), float(deltax), float(nl)

def get_length_from_nodes_thin(diameter, nodes):
    """
    Function to compute the length of a myelinated axon to get the correct number of Nodes of Ranvier
    for A delta thin myelinated models only. Not compatible with myelinated models.

    Attributes
    ----------
    diameter    : float
        diameter of the axon in um
    nodes       : int
        number of nodes in the axon

    Returns
    -------
    length      : float
        lenth of the axon with the correct number of nodes in um
    """
    deltax = 80*(diameter-1)
    return float(math.ceil(deltax*(nodes-1)))

###############################################
## Thin myelinated axons:                    ##
##         specific class for A-delta fibers ##
###############################################
class thin_myelinated(axon):
    """
    Thin Mylineated, A-delta specific, axon class. Automatic refinition of all neuron sections and properties. User-friendly object including model definition
    Inherits from axon class. see axon for further detail.
    """
    def __init__(self, y, z, d, L, model='Extended_Gaines', dt=0.001, node_shift=0,\
        Nseg_per_sec=0, freq=100, freq_min=0, mesh_shape='plateau_sigmoid', alpha_max=0.3,\
        d_lambda=0.1, rec='nodes', v_init=None, md_based_v_init=True, T=None, ID=0, threshold=-40):
        """
        initialisation of a thinnmyelinted (A-delta) axon

        Parameters
        ----------
        y               : float
            y coordinate for the axon, in um
        z               : float
            z coordinate for the axon, in um
        d               : float
            axon diameter, in um. This value has to be greater than 2 for physiological reasons (positive internode length). This object has been designed to be functionnal below 6 um.
        L               : float
            axon length along the x axins, in um
        model           : str
            choice of conductance double-cable based model, possibly:
            'RGK'               : by default, model developped based on ion channels detailed in [1]
            'Extended_Gaines'   : extention of the Gaines sensory model with same doble cable parameters as RGK, see [2] for details on channels
                                        note: the Gaines motor has not been extented as A-delta are supposed to be sensory fibers only
        dt              : float
            computation step for simulations, in ms. By default equal to 1 us
        node_shift      : float
            shift of the first node of Ranvier to zeros, as a fraction of internode length (0<= node_shift < 1)
        Nseg_per_sec    : int
            Number of segment per section in the axon. If set to 0, the number of segment is automatically computed using d-lambda rule and following paramters. If set by user, please use odd numbers
        freq            : float
            Frequency used for the d-lmbda rule, corresponding to the maximum membrane current frequency, by default set to 100 Hz
        freq_min        : float
            Minimal frequency fot the d-lambda rule when using an irregular number of segment along the axon, if set to 0, all sections have the same frequency determined by the previous parameter
        mesh_shape      : str
            Shape of the frequencial distribution for the dlmabda rule along the axon, pick between:
            'pyramidal'         -> min frequencies on both sides and linear increase up to the middle at the maximum frequency
            'sigmoid'           -> same a befor with sigmoid increase instead of linear
            'plateau'           -> sale as pyramidal except the max frequency is holded on a central plateau
            'plateau_sigmoid'   -> same as previous with sigmoid increase
        alpha_max       : float
            Proportion of the axon set to the maximum frequency for plateau shapes, by default set to 0.3
        d_lambda        : float
            value of d-lambda for the dlambda rule
        rec             : str
            recording zones for the membrane potential, eiter:
                'nodes' -> record only at the nodes of Ranvier
            or
                'all' -> all computation points in nodes of Ranvier and over myelin
        v_init          : float
            Initial value of the membrane voltage in mV, set None to get an automatically model attributed value
        md_based_v_init : bool
            If true, the init voltage is decided by the model choice
        T               : gloat
            temperature in C, set None to get an automatically model attributed value
        ID              : int
            axon ID, by default set to 0,
        threshold       : float
            voltage threshold in mV for further spike detection in post-processing, by defautl set to -40mV, see post-processing library for further help

        Note
        ----
        scientific sources for models:
        [1] Tigerholm, J., Poulsen, A. H., Andersen, O. K., and Morch, C. D. (2019). from perception threshold to ion channels—a computational study. Biophysical journal, 117(2), 281-295.
        [2] Gaines, J. L., Finn, K. E., Slopsema, J. P., Heyboer, L. A.,  Polasek, K. H. (2018). A model of motor and sensory axon activation in the median nerve using surface electrical stimulation. Journal of computational neuroscience, 45(1), 29-43.
        """
        super().__init__(y, z, d, L, dt=dt, Nseg_per_sec=Nseg_per_sec,\
            freq=freq, freq_min=freq_min, mesh_shape=mesh_shape, alpha_max=alpha_max, \
            d_lambda=d_lambda, v_init=v_init, T=T, ID=ID, threshold=threshold)
        self.myelinated = True
        self.thin = True
        self.rec = rec
        self.model = model
        self.md_based_v_init = md_based_v_init
        self.node_shift = node_shift

        self.__compute_axon_parameters()

    def __compute_axon_parameters(self):
        """
        generate axon from parameters set by user
        """
        ## Handling v_init
        if self.v_init is None:
            # model driven
            if self.model == 'RGK':
                self.v_init = -64.2
            else:
                self.v_init = -80
        else:
            # user driven
            self.v_init = self.v_init
        ## Handling temperature
        if self.T is None:
            # model driven
            self.T = 37

        ############################################
        ## PARMETERS FOR THE COMPARTIMENTAL MODEL ##
        ############################################
        # compute variable MRG parameters, (usefull also if non MRG models)
        self.axonD, self.nodeD, self.paraD, self.deltax, self.nl = get_Adelta_parameters(self.d)
        # Morphological parameters
        self.nodelength = 2
        self.paralength = 10 # Tigerholm used 5
        self.space_p = 0.002
        self.space_i = 0.004
        self.interlength = (self.deltax-self.nodelength-(2*self.paralength))/6
        # electrical parameters
        self.rhoa = 0.7e6   # Ohm-um
        self.mycm = 0.1     # uF/cm2/lamella membrane
        self.mygm = 0.001   # S/cm2/lamella membrane
        self.Rpn0 = (self.rhoa*.01)/(math.pi*((((self.nodeD/2)+self.space_p)**2)-\
            ((self.nodeD/2)**2)))
        self.Rpn1 = (self.rhoa*.01)/(math.pi*((((self.paraD/2)+self.space_p)**2)-\
            ((self.paraD/2)**2)))
        self.Rpx = (self.rhoa*.01)/(math.pi*((((self.axonD/2)+self.space_i)**2)-\
            ((self.axonD/2)**2)))

        #########################
        ## morphology planning ##
        #########################
        # thin myelinated double cable sequence sequence
        self.MRG_Sequence = ['node', 'MYSA', 'STIN', 'STIN', 'STIN', 'STIN', 'STIN', 'STIN', 'MYSA']
        # basic MRG starts with the node, if needed, adapt the sequence
        if self.node_shift == 0:
            # no Rotation
            self.this_ax_sequence = self.MRG_Sequence
            self.first_section_size = self.nodelength

        elif self.node_shift < (self.paralength)/self.deltax:
            # rotation of less than 1 MYSA
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 1)
            self.first_section_size = self.paralength - (self.node_shift*self.deltax)

        elif self.node_shift < (self.paralength + self.interlength)/self.deltax:
            # rotation of 1 MYSA and less than a STIN
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 2)
            self.first_section_size = self.interlength - (self.node_shift*self.deltax - \
                self.paralength)

        elif self.node_shift < (self.paralength + 2*self.interlength)/self.deltax:
            # rotation of 1 MYSA, 1 STIN and less than a STIN
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 3)
            self.first_section_size = self.interlength - (self.node_shift*self.deltax - \
                self.paralength - self.interlength)

        elif self.node_shift < (self.paralength+ 3*self.interlength)/self.deltax:
            # rotation of 1 MYSA, 2 STIN and less than a STIN
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 4)
            self.first_section_size = self.interlength - (self.node_shift*self.deltax - \
                self.paralength - 2*self.interlength)

        elif self.node_shift < (self.paralength + 4*self.interlength)/self.deltax:
            # rotation of 1 MYSA, 3 STIN and less than a STIN
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 5)
            self.first_section_size = self.interlength - (self.node_shift*self.deltax - \
                self.paralength - 3*self.interlength)

        elif self.node_shift < (self.paralength + 5*self.interlength)/self.deltax:
            # rotation of 1 MYSA, 4 STIN and less than a STIN
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 6)
            self.first_section_size = self.interlength - (self.node_shift*self.deltax - \
                self.paralength - 4*self.interlength)

        elif self.node_shift < (self.paralength + 6*self.interlength)/self.deltax:
            # rotation of 1 MYSA, 5 STIN and less than a STIN
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 7)
            self.first_section_size = self.interlength - (self.node_shift*self.deltax - \
                self.paralength - 5*self.interlength)

        else:
            # rotation of 1 MYSA, 6 STIN and less than a MYSA
            # WARNING FOR DEV : the unprobable case of a node cut in two halfs is not considered...
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 8)
            self.first_section_size = self.deltax*(1-self.node_shift)

        ################
        ## morphology ##
        ################
        self.axonnodes = 0      # number of nodes in the axon
        self.node = []          # list of nodes in the axon
        self.paranodes = 0      # number of MYSA in the axon
        self.MYSA = []          # list of MYSA in the axon
        self.axoninter = 0      # number of STIN in the axon
        self.STIN = []          # list of STIN in the axon

        prov_length = 0
        self.Nsec = 0
        self.last_section_size = 0
        self.last_section_kind = None
        self.axon_path_type = []
        self.axon_path_index = []
        while prov_length < self.L:
            # axon is too short, create a section
            pos_in_sequence = (self.Nsec)%len(self.this_ax_sequence)    # find the current position in the node sequence
            if self.this_ax_sequence[pos_in_sequence] == 'node':
                # you need to create a node
                self.node.append(neuron.h.Section(name='node[%d]' % self.axonnodes))
                self.axon_path_type.append('node')
                self.axon_path_index.append(self.axonnodes)
                self.axonnodes += 1
                self.Nsec += 1
                # increment the prov length and check it is not to much
                if prov_length + self.nodelength >= self.L:
                    # it will be the last section to add to the axon
                    self.last_section_kind = 'node'
                    self.last_section_size = self.L - prov_length
                prov_length += self.nodelength
                # CONNECT THE NODE TO ITS PARENT: a node can only have a MYSA for parent, check it's not the start...
                if self.paranodes != 0:
                    self.node[-1].connect(self.MYSA[-1], 1, 0)

            elif self.this_ax_sequence[pos_in_sequence] == 'MYSA':
                # you need to create a MYSA
                self.MYSA.append(neuron.h.Section(name='MYSA[%d]' % self.paranodes))
                self.axon_path_type.append('MYSA')
                self.axon_path_index.append(self.paranodes)
                self.paranodes += 1
                self.Nsec += 1
                # increment the prov length and check it is not to much
                if prov_length + self.paralength >= self.L:
                    # it will be the last section to add to the axon
                    self.last_section_kind = 'MYSA'
                    self.last_section_size = self.L - prov_length
                prov_length += self.paralength
                # CONNECT THE MYSA TO ITS PARENT: a MYSA can have for parent a node or a TIN, check it's not the start...
                connect_to_type = self.this_ax_sequence[pos_in_sequence - 1] # get parent type
                if connect_to_type == 'node' and self.axonnodes != 0:
                    self.MYSA[-1].connect(self.node[-1], 1, 0)
                if connect_to_type == 'STIN' and self.axoninter != 0:
                    self.MYSA[-1].connect(self.STIN[-1], 1, 0)

            else:
                # you need to create a STIN
                self.STIN.append(neuron.h.Section(name='STIN[%d]' % self.axoninter))
                self.axon_path_type.append('STIN')
                self.axon_path_index.append(self.axoninter)
                self.axoninter += 1
                self.Nsec += 1
                # increment the prov length and check it is not to much
                if prov_length + self.interlength >= self.L:
                    # it will be the last section to add to the axon
                    self.last_section_kind = 'STIN'
                    self.last_section_size = self.L - prov_length
                prov_length += self.interlength
                # CONNECT THE STIN TO ITS PARENT: a STIN can have for parent a STIN od a MYSA, check it's not the start...
                connect_to_type = self.this_ax_sequence[pos_in_sequence - 1]
                if connect_to_type == 'STIN' and self.axoninter > 1:
                    self.STIN[-1].connect(self.STIN[-2], 1, 0)
                if connect_to_type == 'MYSA' and self.paranodes != 0:
                    self.STIN[-1].connect(self.MYSA[-1], 1, 0)

        if self.node == []:
            #logging.warning(
            rise_warning('Warning, thin myelinated axon without node... this can cause latter \
                errors and is maybe unwanted ?\n')

        ####################
        ## programm model ##
        ####################
        self.__set_model(self.model)
        # adjust the length of the first section
        if self.this_ax_sequence[0] == 'node':
            self.node[0].L = self.first_section_size
        elif self.this_ax_sequence[0] == 'MYSA':
            self.MYSA[0].L = self.first_section_size
        else: # should be a STIN
            self.STIN[0].L = self.first_section_size
        # adjust the length if the last section
        if self.last_section_kind == 'node':
            self.node[-1].L = self.last_section_size
        elif self.last_section_kind == 'MYSA':
            self.MYSA[-1].L = self.last_section_size
        else: # should be a STIN
            self.STIN[-1].L = self.last_section_size
        # define the geometry of the axon
        self._axon__define_shape()
        # define the number of segments
        self.__set_Nseg()
        # get nodes positions
        self.__get_seg_positions()
        self.__get_rec_positions()


    def save(self, save=False, fname='axon.json', extracel_context=False, intracel_context=False, rec_context=False):
        """
        Return axon as dictionary and eventually save it as json file

        Parameters
        ----------
        save    : bool
            if True, save in json files
        fname   : str
            Path and Name of the saving file, by default 'axon.json'

        Returns
        -------
        ax_dic : dict
            dictionary containing all information
        """
        ax_dic = super().save(extracel_context=extracel_context, intracel_context=intracel_context, rec_context=rec_context)


        ax_dic['myelinated'] = self.myelinated
        ax_dic['thin'] = self.thin
        ax_dic['rec'] = self.rec
        ax_dic['node_shift'] = self.node_shift
        ax_dic['md_based_v_init'] = self.md_based_v_init
        ax_dic['model'] = self.model

        if save:
            json_dump(ax_dic, fname)
        return ax_dic

    def load(self, data, extracel_context=False, intracel_context=False, rec_context=False):
        """
        Load all axon properties from a dictionary or a json file

        Parameters
        ----------
        data    : str or dict
            json file path or dictionary containing axon information
        """
        if type(data) == str:
            ax_dic = json_load(data)
        else: 
            ax_dic = data
        super().load(data, extracel_context=extracel_context, intracel_context=intracel_context, rec_context=rec_context)

        self.myelinated = ax_dic['myelinated']
        self.thin = ax_dic['thin']
        self.rec = ax_dic['rec']
        self.node_shift = ax_dic['node_shift']
        self.md_based_v_init = ax_dic['md_based_v_init']
        if ax_dic['model'] in myelinated_models:
            self.model = ax_dic['model']
        else:
            self.model = 'Extended_Gaines'
        self.__compute_axon_parameters()
        if intracel_context:
            for i in range(len(ax_dic['intra_current_stim_positions'])):
                position = ax_dic['intra_current_stim_positions'][i]
                stim_start = ax_dic['intra_current_stim_starts'][i]
                duration = ax_dic['intra_current_stim_durations'][i]
                amplitude = ax_dic['intra_current_stim_amplitudes'][i]
                self.insert_I_Clamp(position/self.L, stim_start, duration, amplitude)
            if ax_dic['intra_voltage_stim_stimulus'] is not None:
                position = ax_dic['intra_voltage_stim_position'][0]
                stim = stimulus()
                stim.load(ax_dic['intra_voltage_stim_stimulus'])
                self.insert_V_Clamp(position/self.L, stim)


    def __set_model(self, model):
        """
        Set the double cable model. For internal use only.

        Parameters
        ----------
        model           : str
            choice of conductance double-cable based model, possibly:
                'RGK'               : specific A-delta model
                'Extended_Gaines'   : extension of the Gaines sensory model to lower diameter values
        """
        # NB: the remaining commented values correspond to an attempt of MRG extension
        # This does not work... stimulation causes a double spike, probabely caused by lack of Potassium channel on paranodes
        # comment to keep for demos only...
        for n in self.node:
            n.nseg = 1
            n.diam = self.nodeD
            n.L = self.nodelength
            if model == 'Extended_Gaines':
                #n.insert('axnode')
                n.Ra = self.rhoa/10000
                n.cm = 2
                n.insert('node_sensory')
            else: #RGK
                n.Ra = 130
                n.cm = 1
                n.insert('nav1p9')
                n.gbar_nav1p9 = 1.1e-3
                n.celsiusT_nav1p9 = self.T
                n.insert('nax')
                n.gbar_nax = 1.45e-1
                n.insert('ks')
                n.gbar_ks = 2.10e-3
                n.v = self.v_init
                n.insert('pas')
                n.g_pas = 1e-7
                n.e_pas = -60
            n.insert('extracellular')
            n.xraxial[0] = self.Rpn0
            n.xg[0] = 1e10
            n.xc[0] = 0
        for m in self.MYSA:
            m.nseg = 1
            m.diam = self.d
            m.L = self.paralength
            m.cm = 2*self.paraD/self.d
            if model == 'Extended_Gaines':
                m.Ra = self.rhoa*(1/(self.paraD/self.d)**2)/10000
                #m.insert('pas')
                #m.g_pas = 0.001*self.paraD/self.d
                #m.e_pas = -80
                m.insert('mysa_sensory')
            else:
                m.Ra = 130
                m.insert('kdrTiger')
                m.celsiusT_kdrTiger = self.T
                m.gbar_kdrTiger = 4.80e-3
                m.insert('h')
                m.gbar_h = 1.52e-4
                m.celsiusT_h = self.T
                m.insert('kaslow')
                m.gbar_kaslow = 3.0e-3
                m.v = self.v_init
                m.insert('pas')
                m.g_pas = 1e-7
                m.e_pas = -60
            m.insert('extracellular')
            m.xraxial[0] = self.Rpn1
            m.xg[0] = self.mygm/(self.nl*2)
            m.xc[0] = self.mycm/(self.nl*2)
        for s in self.STIN:
            s.nseg = 1
            s.diam = self.d
            s.L = self.interlength
            s.cm = 2*self.axonD/self.d
            if self.model == 'Extended_Gaines':
                s.Ra = self.rhoa*(1/(self.axonD/self.d)**2)/10000
                s.insert('stin_sensory')
            else:
                s.Ra = 130
                s.insert('pas')
                s.g_pas = 1e-7#0.0001*self.axonD/self.d
                s.e_pas = -60
            s.insert('extracellular')
            s.xraxial[0] = self.Rpx
            s.xg[0] = self.mygm/(self.nl*2)
            s.xc[0] = self.mycm/(self.nl*2)

    def __set_Nseg(self):
        """
        Set the number of segments automatically acording initialization. For internal use only.
        """
        if self.Nseg_per_sec != 0:
            # all sections will have a fixed number of segment chosen by user
            for n in self.node:
                n.nseg = self.Nseg_per_sec
            for m in self.MYSA:
                m.nseg = self.Nseg_per_sec
            for s in self.STIN:
                s.nseg = self.Nseg_per_sec
            self.node_Nseg = self.axonnodes * self.Nseg_per_sec
            self.MYSA_Nseg = self.paranodes * self.Nseg_per_sec
            self.STIN_Nseg = self.axoninter * self.Nseg_per_sec
        else:
            # here comes the dlambda rule
            self.node_Nseg = 0
            self.MYSA_Nseg = 0
            self.STIN_Nseg = 0
            if self.freq_min == 0:
                # uniform meshing
                for n in self.node:
                    Nseg = d_lambda_rule(n.L, self.d_lambda, self.freq, n)
                    n.nseg = Nseg
                    self.node_Nseg += Nseg
                for m in self.MYSA:
                    Nseg = Nseg = d_lambda_rule(m.L, self.d_lambda, self.freq, m)
                    m.nseg = Nseg
                    self.MYSA_Nseg += Nseg
                for s in self.STIN:
                    Nseg = Nseg = d_lambda_rule(s.L, self.d_lambda, self.freq, s)
                    s.nseg = Nseg
                    self.STIN_Nseg += Nseg
            else:
                # non-uniform meshing
                freqs = create_Nseg_freq_shape(self.Nsec, self.mesh_shape, self.freq, \
                    self.freq_min, self.alpha_max)
                for k in range(len(self.axon_path_type)):
                    sec_type = self.axon_path_type[k]
                    sec_index = self.axon_path_index[k]
                    if sec_type == 'node':
                        Nseg = d_lambda_rule(self.node[sec_index].L, self.d_lambda, self.freq, \
                            self.node[sec_index])
                        self.node[sec_index].nseg = Nseg
                        self.node_Nseg += Nseg
                    elif sec_type == 'MYSA':
                        Nseg = d_lambda_rule(self.MYSA[sec_index].L, self.d_lambda, self.freq, \
                            self.MYSA[sec_index])
                        self.MYSA[sec_index].nseg = Nseg
                        self.MYSA_Nseg += Nseg
                    else: # should be STIN
                        Nseg = d_lambda_rule(self.STIN[sec_index].L, self.d_lambda, self.freq, \
                            self.STIN[sec_index])
                        self.STIN[sec_index].nseg = Nseg
                        self.STIN_Nseg += Nseg
        self.Nseg = self.node_Nseg + self.MYSA_Nseg + self.STIN_Nseg

    def __get_seg_positions(self):
        """
        Get segment positions, for internal use only.
        """
        x_offset = 0
        x = []
        x_nodes = []
        nodes_index = []
        self.rec_position_list = []
        for k in range(len(self.axon_path_type)):
            sec_type = self.axon_path_type[k]
            sec_index = self.axon_path_index[k]
            self.rec_position_list.append([])
            if sec_type == 'node':
                x_nodes.append(x_offset + self.node[sec_index].L/2)
                for seg in self.node[sec_index].allseg():
                    if x == []:
                        x.append(seg.x*(self.node[sec_index].L) + x_offset)
                        nodes_index.append(0)
                        self.rec_position_list[-1].append(seg.x)
                    else:
                        x_seg = seg.x*(self.node[sec_index].L) + x_offset
                        if x_seg != x[-1]:
                            x.append(x_seg)
                            nodes_index.append(len(x)-1)
                            self.rec_position_list[-1].append(seg.x)
                x_offset += self.node[sec_index].L
            elif sec_type == 'MYSA':
                for seg in self.MYSA[sec_index].allseg():
                    if x == []:
                        x.append(seg.x*(self.MYSA[sec_index].L) + x_offset)
                        self.rec_position_list[-1].append(seg.x)
                    else:
                        x_seg = seg.x*(self.MYSA[sec_index].L) + x_offset
                        if x_seg != x[-1]:
                            x.append(x_seg)
                            self.rec_position_list[-1].append(seg.x)
                x_offset += self.MYSA[sec_index].L
            else: # should be STIN
                for seg in self.STIN[sec_index].allseg():
                    if x == []:
                        x.append(seg.x*(self.STIN[sec_index].L) + x_offset)
                        self.rec_position_list[-1].append(seg.x)
                    else:
                        x_seg = seg.x*(self.STIN[sec_index].L) + x_offset
                        if x_seg != x[-1]:
                            x.append(x_seg)
                            self.rec_position_list[-1].append(seg.x)
                x_offset += self.STIN[sec_index].L
        self.x = np.asarray(x)
        self.x_nodes = np.asarray(x_nodes)
        self.node_index = np.asarray(nodes_index)

    def __get_rec_positions(self):
        """
        Get the position of points with voltage recording. For internal use only.
        """
        if self.rec == 'nodes':
            self.x_rec = self.x_nodes
        else:
            self.x_rec = self.x

    ###############################
    ## Intracellular stimulation ##
    ###############################
    def insert_I_Clamp_node(self, index, t_start, duration, amplitude):
        """
        Insert a IC clamp stimulation on a Ranvier node at its midd point position

        Parameters
        ----------
        index       : int
            node number of the node to stimulate
        t_start     : float
            starting time (ms)
        duration    : float
            duration of the pulse(ms)
        amplitude   : float
            amplitude of the pulse (nA)
        """
        # add the stimulation to the axon
        self.intra_current_stim.append(neuron.h.IClamp(0.5, sec=self.node[index]))
        # modify the stimulation parameters
        self.intra_current_stim[-1].delay = t_start
        self.intra_current_stim[-1].dur = duration
        self.intra_current_stim[-1].amp = amplitude
        # save the stimulation parameter for results
        self.intra_current_stim_positions.append(self.x_nodes[index])
        self.intra_current_stim_starts.append(t_start)
        self.intra_current_stim_durations.append(duration)
        self.intra_current_stim_amplitudes.append(amplitude)

    def insert_I_Clamp(self, position, t_start, duration, amplitude):
        """
        Insert a IC clamp stimulation at the midd point of the nearest node to the specified position

        Parameters
        ----------
        position    : float
            relative position over the axon
        t_start     : float
            starting time (ms)
        duration    :
            duration of the pulse(ms)
        amplitude   :
            amplitude of the pulse (nA)
        """
        # adapt position to the number of sections
        index = round((position * (self.axonnodes - 1)+0.5))
        self.insert_I_Clamp_node(index, t_start, duration, amplitude)

    def insert_V_Clamp_node(self, index, stimulus):
        """
        Insert a V clamp stimulation

        Parameters
        ----------
        index       : int
            node number of the node to stimulate
        stimulus    : stimulus object
            stimulus for the clamp, see Stimulus.py for more information
        """
        # add the stimulation to the axon
        self.intra_voltage_stim = neuron.h.VClamp(0.5, sec=self.node[index])
        # save the stimulation parameter for results
        self.intra_voltage_stim_position.append(self.x_nodes[index])
        # save the stimulus for later use
        self.intra_voltage_stim_stimulus = stimulus
        # set fake duration
        self.intra_voltage_stim.dur[0] = 1e9

    def insert_V_Clamp(self, position, stimulus):
        """
        Insert a V clamp stimulation at the midd point of the nearest node to the specified position

        Parameters
        ----------
        position    : float
            relative position over the axon
        stimulus    : stimulus object
            stimulus for the clamp, see Stimulus.py for more information
        """
        # adapt position to the number of sections
        index = round((position * (self.axonnodes - 1)+0.5))
        self.insert_V_Clamp_node(index, stimulus)

    ##############################
    ## Result recording methods ##
    ##############################
    def _set_recorders_with_key(self, reclist, key=None, single_mod=True, key_node=None,\
        key_MYSA=None, key_FLUT=None, key_STIN=None):
        """
        To automate the methods set_recorder. For internal use only.
        Parameters
        ----------
        reclist     : neuron.h.List
            List in witch the reccorders should be saved
        key         : str
            Name of the key which should be recorded, if None result of the recording set to [0],
            by default None
        single_mod  : bool
            if True key will be used for all the section, else specific key will be used in each
            section, by default True
        key_node    : str
            Name of the key which should be recorded in node, if None result of the recording set
            to [0], by default None
        key_MYSA    : str
            Name of the key which should be recorded in MYSA, if None result of the recording set
            to [0], by default None
        key_FLUT    : str
            Name of the key which should be recorded in FLUT, if None result of the recording set
            to [0], by default None
        key_STIN    : str
            Name of the key which should be recorded in STIN, if None result of the recording set
            to [0], by default None
        """
        if single_mod:
            key_node, key_MYSA, key_FLUT, key_STIN = key, key, key, key

        if self.rec == 'nodes':

            # recording only on middle of all nodes
            for n in self.node:
                if key_node is not None:
                    rec = neuron.h.Vector().record(getattr(n(0.5), key_node), sec=n)
                else:
                    rec = neuron.h.Vector([0])
                reclist.append(rec)
        else:
            # recording on all segments
            for k in range(len(self.axon_path_type)):
                sec_type = self.axon_path_type[k]
                sec_index = self.axon_path_index[k]
                for position in self.rec_position_list[k]:
                    if sec_type == 'node':
                        if key_node is not None:
                            rec = neuron.h.Vector().record(getattr(self.node[sec_index](position), key_node),\
                                sec=self.node[sec_index])
                        else:
                            rec = neuron.h.Vector([0])
                        reclist.append(rec)
                    elif sec_type == 'MYSA':
                        if key_MYSA is not None:
                            rec = neuron.h.Vector().record(getattr(self.MYSA[sec_index](position), key_MYSA), \
                                sec=self.MYSA[sec_index])
                        else:
                            rec = neuron.h.Vector([0])                            
                        reclist.append(rec)
                    elif sec_type == 'FLUT':
                        if key_FLUT is not None:
                            rec = neuron.h.Vector().record(getattr(self.FLUT[sec_index](position), key_FLUT), \
                                sec=self.FLUT[sec_index])
                        else:
                            rec = neuron.h.Vector([0]) 
                        reclist.append(rec)
                    else: # should be STIN
                        if key_STIN is not None:
                            rec = neuron.h.Vector().record(getattr(self.STIN[sec_index](position), key_STIN), \
                                sec=self.STIN[sec_index])
                        else:
                            rec = neuron.h.Vector([0]) 
                        reclist.append(rec)

    def _get_recorders_from_list(self, reclist):
        """
        Convert reclist in np.array To automate methods set_recorder. For internal use only.
        Parameters
        ----------
        reclist     : neuron.h.List
            List in witch the reccorders are saved
        
        Returns
        -------
        val         : np.array
            array of every recorded value for all rec point and time
        """
        if self.rec == 'nodes':
            dim = (self.axonnodes, self.t_len)
        else:
            dim = (len(self.x_rec), self.t_len)

        val = np.zeros(dim)
        for k in range(dim[0]):
            val[k, :] = np.asarray(reclist[k])
        return val

    def _get_var_from_mod(self, key_node=None, key_MYSA=None, key_FLUT=None, key_STIN=None):
        """
        return a column with value in every recording point of a constant from a mod. For internal use only.
        """
        if self.rec == 'nodes':
            val = np.zeros((len(self.node),1))
            if key_node is not None:
                val[:,0] = getattr(self.node[0](0.5), key_node)[0]
        else:
            val = np.zeros((len(self.x_rec), 1))
            i=-1
            for k in range(len(self.axon_path_type)):
                sec_type = self.axon_path_type[k]
                sec_index = self.axon_path_index[k]
                for position in self.rec_position_list[k]:
                    i += 1
                    if sec_type == 'node':
                        if key_node is not None:
                            val[i,0] = getattr(self.node[sec_index](position), key_node)[0]
                    elif sec_type == 'MYSA':
                        if key_MYSA is not None:
                            val[i,0] = getattr(self.MYSA[sec_index](position), key_MYSA)[0]
                    elif sec_type == 'FLUT':
                        if key_FLUT is not None:
                            val[i,0] = getattr(self.FLUT[sec_index](position), key_FLUT)[0]
                    else: # should be STIN
                        if key_STIN is not None:
                            val[i,0] = getattr(self.STIN[sec_index](position), key_STIN)[0]
        return val

    def set_membrane_voltage_recorders(self):
        """
        Prepare the membrane voltage recording. For internal use only.
        """
        self.vreclist = neuron.h.List()
        key = '_ref_v'
        self._set_recorders_with_key(self.vreclist, key)

    def get_membrane_voltage(self):
        """
        get the membrane voltage at the end of simulation. For internal use only.
        """
        return self._get_recorders_from_list(self.vreclist)

    def set_membrane_current_recorders(self):
        """
        Prepare the membrane current recording. For internal use only.
        """
        self.ireclist = neuron.h.List()
        key = '_ref_i_membrane'
        self._set_recorders_with_key(self.ireclist, key)

    def get_membrane_current(self):
        """
        get the membrane current at the end of simulation. For internal use only.
        """
        return self._get_recorders_from_list(self.ireclist)


    def set_ionic_current_recorders(self):
        """
        Prepare the ionic current recording. For internal use only.
        """
        if self.model == 'Extended_Gaines':
            self.gaines_ina_reclist = neuron.h.List()
            self.gaines_inap_reclist = neuron.h.List()
            self.gaines_ik_reclist = neuron.h.List()
            self.gaines_ikf_reclist = neuron.h.List()
            self.gaines_il_reclist = neuron.h.List()
            for n in self.node:
                gaines_ina = neuron.h.Vector().record(n(0.5)._ref_ina_node_sensory, sec=n)
                gaines_inap = neuron.h.Vector().record(n(0.5)._ref_inap_node_sensory, sec=n)
                gaines_ik = neuron.h.Vector().record(n(0.5)._ref_ik_node_sensory, sec=n)
                gaines_ikf = neuron.h.Vector().record(n(0.5)._ref_ikf_node_sensory, sec=n)
                gaines_il = neuron.h.Vector().record(n(0.5)._ref_il_node_sensory, sec=n)
                self.gaines_ina_reclist.append(gaines_ina)
                self.gaines_inap_reclist.append(gaines_inap)
                self.gaines_ik_reclist.append(gaines_ik)
                self.gaines_ikf_reclist.append(gaines_ikf)
                self.gaines_il_reclist.append(gaines_il)
        else:
            self.RGK_ina_reclist = neuron.h.List()
            self.RGK_ik_reclist = neuron.h.List()
            self.RGK_ipas_reclist = neuron.h.List()
            for n in self.node:
                RGK_ina = neuron.h.Vector().record(n(0.5)._ref_nai, sec=n)
                RGK_ik = neuron.h.Vector().record(n(0.5)._ref_ki, sec=n)
                RGK_ipas = neuron.h.Vector().record(n(0.5)._ref_i_pas, sec=n)
                self.RGK_ina_reclist.append(RGK_ina)
                self.RGK_ik_reclist.append(RGK_ik)
                self.RGK_ipas_reclist.append(RGK_ipas)

    def get_ionic_current(self):
        """
        get the ionic currents at the end of simulation. For internal use only.
        """
        if self.model == 'Extended_Gaines':
            gaines_ina_ax = np.zeros((self.axonnodes, self.t_len))
            gaines_inap_ax = np.zeros((self.axonnodes, self.t_len))
            gaines_ik_ax = np.zeros((self.axonnodes, self.t_len))
            gaines_ikf_ax = np.zeros((self.axonnodes, self.t_len))
            gaines_il_ax = np.zeros((self.axonnodes, self.t_len))
            for k in range(self.axonnodes):
                gaines_ina_ax[k, :] = np.asarray(self.gaines_ina_reclist[k])
                gaines_inap_ax[k, :] = np.asarray(self.gaines_inap_reclist[k])
                gaines_ik_ax[k, :] = np.asarray(self.gaines_ik_reclist[k])
                gaines_ikf_ax[k, :] = np.asarray(self.gaines_ikf_reclist[k])
                gaines_il_ax[k, :] = np.asarray(self.gaines_il_reclist[k])
            results = [gaines_ina_ax, gaines_inap_ax, gaines_ik_ax, gaines_ikf_ax, gaines_il_ax]
        else:
            RGK_ina_ax = np.zeros((self.axonnodes, self.t_len))
            RGK_ik_ax = np.zeros((self.axonnodes, self.t_len))
            RGK_ipas_ax = np.zeros((self.axonnodes, self.t_len))
            for k in range(self.axonnodes):
                RGK_ina_ax[k, :] = np.asarray(self.RGK_ina_reclist[k])
                RGK_ik_ax[k, :] = np.asarray(self.RGK_ik_reclist[k])
                RGK_ipas_ax[k, :] = np.asarray(self.RGK_ipas_reclist[k])
            results = [RGK_ina_ax, RGK_ik_ax, RGK_ipas_ax]
        return results

    def set_particules_values_recorders(self):
        """
        Prepare the particules values recording. For internal use only.
        """
        if self.model == 'Extended_Gaines':
            self.gaines_mreclist = neuron.h.List()
            self.gaines_mpreclist = neuron.h.List()
            self.gaines_sreclist = neuron.h.List()
            self.gaines_hreclist = neuron.h.List()
            self.gaines_nreclist = neuron.h.List()
            for n in self.node:
                gaines_m = neuron.h.Vector().record(n(0.5)._ref_m_node_sensory, sec=n)
                gaines_mp = neuron.h.Vector().record(n(0.5)._ref_mp_node_sensory, sec=n)
                gaines_s = neuron.h.Vector().record(n(0.5)._ref_s_node_sensory, sec=n)
                gaines_h = neuron.h.Vector().record(n(0.5)._ref_h_node_sensory, sec=n)
                gaines_n = neuron.h.Vector().record(n(0.5)._ref_n_node_sensory, sec=n)
                self.gaines_mreclist.append(gaines_m)
                self.gaines_mpreclist.append(gaines_mp)
                self.gaines_sreclist.append(gaines_s)
                self.gaines_hreclist.append(gaines_h)
                self.gaines_nreclist.append(gaines_n)
        else:
            self.RGK_m_nav1p9_reclist = neuron.h.List()
            self.RGK_h_nav1p9_reclist = neuron.h.List()
            self.RGK_s_nav1p9_reclist = neuron.h.List()
            self.RGK_m_nax_reclist = neuron.h.List()
            self.RGK_h_nax_reclist = neuron.h.List()
            self.RGK_ns_ks_reclist = neuron.h.List()
            self.RGK_nf_ks_reclist = neuron.h.List()
            for n in self.node:
                # Nav1p9
                m_nav1p9 = neuron.h.Vector().record(n(0.5)._ref_m_nav1p9, sec=n)
                h_nav1p9 = neuron.h.Vector().record(n(0.5)._ref_h_nav1p9, sec=n)
                s_nav1p9 = neuron.h.Vector().record(n(0.5)._ref_s_nav1p9, sec=n)
                # Nax
                m_nax = neuron.h.Vector().record(n(0.5)._ref_m_nax, sec=n)
                h_nax = neuron.h.Vector().record(n(0.5)._ref_h_nax, sec=n)
                # Ks
                ns_ks = neuron.h.Vector().record(n(0.5)._ref_ns_ks, sec=n)
                nf_ks = neuron.h.Vector().record(n(0.5)._ref_nf_ks, sec=n)
                self.RGK_m_nav1p9_reclist.append(m_nav1p9)
                self.RGK_h_nav1p9_reclist.append(h_nav1p9)
                self.RGK_s_nav1p9_reclist.append(s_nav1p9)
                self.RGK_m_nax_reclist.append(m_nax)
                self.RGK_h_nax_reclist.append(h_nax)
                self.RGK_ns_ks_reclist.append(ns_ks)
                self.RGK_nf_ks_reclist.append(nf_ks)

    def get_particles_values(self):
        """
        get the particules values at the end of simulation. For internal use only.
        """
        if self.model == 'Extended_Gaines':
            gaines_m_ax = np.zeros((self.axonnodes, self.t_len))
            gaines_mp_ax = np.zeros((self.axonnodes, self.t_len))
            gaines_s_ax = np.zeros((self.axonnodes, self.t_len))
            gaines_h_ax = np.zeros((self.axonnodes, self.t_len))
            gaines_n_ax = np.zeros((self.axonnodes, self.t_len))
            for k in range(self.axonnodes):
                gaines_m_ax[k, :] = np.asarray(self.gaines_mreclist[k])
                gaines_mp_ax[k, :] = np.asarray(self.gaines_mpreclist[k])
                gaines_s_ax[k, :] = np.asarray(self.gaines_sreclist[k])
                gaines_h_ax[k, :] = np.asarray(self.gaines_hreclist[k])
                gaines_n_ax[k, :] = np.asarray(self.gaines_nreclist[k])
            results = [gaines_m_ax, gaines_mp_ax, gaines_s_ax, gaines_h_ax, gaines_n_ax]
        else:
            RGK_m_nav1p9_ax = np.zeros((self.axonnodes, self.t_len))
            RGK_h_nav1p9_ax = np.zeros((self.axonnodes, self.t_len))
            RGK_s_nav1p9_ax = np.zeros((self.axonnodes, self.t_len))
            RGK_m_nax_ax = np.zeros((self.axonnodes, self.t_len))
            RGK_h_nax_ax = np.zeros((self.axonnodes, self.t_len))
            RGK_ns_ks_ax = np.zeros((self.axonnodes, self.t_len))
            RGK_nf_ks_ax = np.zeros((self.axonnodes, self.t_len))
            for k in range(self.axonnodes):
                RGK_m_nav1p9_ax[k, :] = np.asarray(self.RGK_m_nav1p9_reclist[k])
                RGK_h_nav1p9_ax[k, :] = np.asarray(self.RGK_h_nav1p9_reclist[k])
                RGK_s_nav1p9_ax[k, :] = np.asarray(self.RGK_s_nav1p9_reclist[k])
                RGK_m_nax_ax[k, :] = np.asarray(self.RGK_m_nax_reclist[k])
                RGK_h_nax_ax[k, :] = np.asarray(self.RGK_h_nax_reclist[k])
                RGK_ns_ks_ax[k, :] = np.asarray(self.RGK_ns_ks_reclist[k])
                RGK_nf_ks_ax[k, :] = np.asarray(self.RGK_nf_ks_reclist[k])
            results = [RGK_m_nav1p9_ax, RGK_h_nav1p9_ax, RGK_s_nav1p9_ax, RGK_m_nax_ax,\
                RGK_h_nax_ax, RGK_ns_ks_ax, RGK_nf_ks_ax]
        return results



"""
Access and modify NRV Parameters
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import configparser
import os
from numpy import pi


class nrv_parameters():
    def __init__(self):
        """
        """
        self.dir_path = os.environ['NRVPATH'] + '/_misc'
        self.config_fname = self.dir_path + '/NRV.ini'

        self.load()

    def save(self):
        with open(self.config_fname, 'w') as configfile:
            self.machine_config.write(configfile)
    
    def load(self, fname=None):
        if fname is None:
            fname = self.config_fname

        self.machine_config = configparser.ConfigParser()
        self.machine_config.read(self.config_fname)

        #GMSH
        self.GMSH_Ncores = int(self.machine_config.get('GMSH', 'GMSH_CPU'))
        self.GMSH_Status = self.machine_config.get('GMSH', 'GMSH_STATUS') == 'True'
        #LOG
        self.LOG_Status = self.machine_config.get('LOG', 'LOG_STATUS') == 'True'
        self.VERBOSITY_LEVEL = int(self.machine_config.get('LOG', 'VERBOSITY_LEVEL'))

    def get_nrv_verbosity(self):
        """
        get general verbosity level
        """
        return self.VERBOSITY_LEVEL


    def set_nrv_verbosity(self, i):
        """
        set general verbosity level
        Parameters
        ----------
        O: None
        1: + Error
        2: + Warning
        3: + Info
        4: + Debug
        NB: to add Debug verbosity to the log VERBOSITY_LEVEL has to be set to 4
        """
        self.VERBOSITY_LEVEL = i


    def get_gmsh_ncore(self):
        """
        get gmsh core number
        """
        return self.GMSH_Ncores

    def set_gmsh_ncore(self,n):
        """
        set gmsh core number
        """
        self.GMSH_Ncores = n

##########################################################################
########################### Parameter singleton ##########################
##########################################################################

parameters = nrv_parameters()



"""
NRV-FEM
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq
(c) ETIS - University Cergy-Pontoise - CNRS
"""
from abc import abstractmethod
from ...backend.log_interface import rise_error, rise_warning, pass_info
from ...backend.NRV_Class import NRV_class
import time

###############
## Constants ##
###############
fem_verbose = True

###################
## Model classes ##
###################
class FEM_model(NRV_class):
    """
    A generic class for Finite Element models
    """
    @abstractmethod
    def __init__(self, Ncore=None):
        """
        Creates a instance of FEM_model

        Parameters
        ----------
        Ncore   : int
            number of cores for computation. If None is specified, this number is taken from the NRV2.ini configuration file. Byu default set to None
        """
        super().__init__()
        self.Ncore = Ncore
        self.type = 'FEM'

        # Timmers
        self.meshing_timer = 0
        self.preparing_timer = 0
        self.solving_timer = 0
        self.access_res_timer = 0

        # Flags
        self.is_meshed = False
        self.is_computed = False

    
    def get_timers(self, verbose=False):
        """
        
        """
        pass_info("mesh done in " + str(self.meshing_timer) + " s")
        pass_info("simulation prepared in " + str(self.preparing_timer) + " s")
        pass_info("simulation solved in " + str(self.solving_timer) + " s")
        pass_info("Time spent to access results " + str(self.access_res_timer) + " s")
        total_timer = self.meshing_timer + self.preparing_timer + self.solving_timer + self.access_res_timer
        pass_info("total duration " + str(total_timer) + " s")
        return self.meshing_timer, self.preparing_timer, self.solving_timer, total_timer



nrv.rasterize(sim_results,'V_mem')


list_keys = ['ID', 'L', 'V_mem_raster_position', 'V_mem_raster_x_position', 'V_mem_raster_time_index', \
'V_mem_raster_time', 'myelinateds', 'intra_stim_starts', 'intra_stim_positions']

removable_keys = []
for key in sim_results:
    if not key in list_keys:
        removable_keys += [key]

for key in removable_keys:
    nrv.remove_key(sim_results,key, verbose=verbose)




from ..backend.log_interface import rise_error, rise_warning, pass_info
from ..backend.file_handler import json_dump
from ..backend.NRV_Class import NRV_class

from abc import ABCMeta, abstractmethod

import numpy as np
#import pyswarms


class Optimizer(NRV_class, metaclass=ABCMeta):
  @abstractmethod
  def __init__(self, method=None):
    self._method = method

class PSO_optimizer(Optimizer):
  def __init__(self):
    super().__init__('PSO')
    pass



"""
NRV-FEM
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import numpy as np
import configparser
import os

from ...utils.units import V
from ...backend.file_handler import rmv_ext
from .FEM import *

from .mesh_creator.NerveMshCreator import *
from .fenics_utils.FEMSimulation import *
from.fenics_utils.SimResult import *

# built in FENICS models
dir_path = os.environ['NRVPATH'] + '/_misc'
#material_library = os.listdir(dir_path+'/fenics_templates/')

###############
## Constants ##
###############
machine_config = configparser.ConfigParser()
config_fname = dir_path + '/NRV.ini'
machine_config.read(config_fname)

FENICS_Ncores = machine_config.get('FENICS', 'FENICS_CPU')
FENICS_Status = machine_config.get('FENICS', 'FENICS_STATUS') == 'True'

fem_verbose = True


class FENICS_model(FEM_model):
    """
    A class for FENICS Finite Element Models, inherits from FEM_model.
    """
    def __init__(self, fname=None, Ncore=None, handle_server=False, elem=None, comm=MPI.COMM_SELF,\
        rank=0):
        """
        Creates a instance of a FENICS Finite Element Model object.

        Parameters
        ----------
        fname   : str
            path to the mesh file (.msh) file
        Ncore   : int
            number of FENICS cores for computation. If None is specified, this number is taken from the NRV2.ini configuration file. Byu default set to None
        handle_server   : bool
            if True, the instantiation creates the server, else a external server is used. Usefull for multiple cells sharing the same model
        """
        super().__init__(Ncore=Ncore)
        self.type = 'FENICS'
        
        # Default paramerters
        self.L = 5000
        self.y_c = 0
        self.z_c = 0
        self.Outer_D = 5       #mm
        self.Nerve_D = 250     #um
        self.N_fascicle = 0
        self.fascicles = {}
        self.Perineurium_thickness={}
        self.N_electrode = 0
        self.electrodes = {}
        self.Istim = 1e-3        #A
        self.jstims = []
        self.j_electrode = {}

     
        self.Outer_mat = "saline"
        self.Epineurium_mat = "epineurium"
        self.Endoneurium_mat = "endoneurium_ranck"
        self.Perineurium_mat = "perineurium"
        self.Electrodes_mat = 1#"platinum"

        self.default_fascicle = {"D":200, "y_c":0, "z_c":0, "res":20}
        self.default_electrode = {"elec_type":"LIFE", "x_c":self.L/2, "y_c":0, "z_c":0, "length":1000, "D":25, "res":3}

        # Mesh 
        self.mesh_file = fname
        self.mesh = None
        self.elem = elem
        if elem is None:
            self.elem = ('Lagrange', 2)

        # FEM 
        self.sim = None
        self.sim_res = []

        # Mcore attributes
        self.comm = comm
        self.rank = rank

        # Status
        self.mesh_file_status = not self.mesh_file is None
        self.is_sim_ready = False

        if not self.mesh_file_status:
            self.mesh = NerveMshCreator(Length=self.L, Outer_D=self.Outer_D, Nerve_D=self.Nerve_D,\
                y_c=self.y_c, z_c=self.z_c)




    def save(self, fname, type='msh'):
        """
        Save the changes to the model file. (Avoid for the overal weight of the package)
        """
        if self.is_computed:
            save_sim_res_list(self.sim_res, fname)

        elif self.is_meshed: 
            self.mesh_file = rmv_ext(fname)
            self.mesh.save(fname)


    #############################
    ## Access model parameters ##
    #############################
    def get_parameters(self):
        """
        Get the  all the parameters in the model as a python dictionary.

        Returns
        -------
        list    : dict
            all parameters as dictionnaries in a list, names a keys, with corresponding values
        """
        param = {}
        param['L'] = self.L
        param['y_c'] = self.y_c
        param['z_c'] = self.z_c
        param['Outer_D'] = self.Outer_D
        param['Nerve_D'] = self.Nerve_D
        param['N_fascicle'] = self.N_fascicle
        param['fascicles'] = self.fascicles
        param['Perineurium_thickness'] = self.Perineurium_thickness
        param['N_electrode'] = self.N_electrode
        param['electrodes'] = self.electrodes
        
        param['Outer_mat'] = self.Outer_mat
        param['Epineurium_mat'] = self.Epineurium_mat
        param['Endoneurium_mat'] = self.Endoneurium_mat
        param['Perineurium_mat'] = self.Perineurium_mat
        param['Perineurium_mat'] = self.Perineurium_mat
        
        param['Istim'] = self.Istim 
        return param

    def __update_parameters(self):
        """
        Internal use only: updates all the parameters from the mesh
        """   
        param = self.mesh.get_parameters()
        self.L = param['L']
        self.y_c = param['y_c']
        self.z_c = param['z_c']
        self.Outer_D = param['Outer_D']
        self.Nerve_D = param['Nerve_D']
        self.N_fascicle = param['N_fascicle']
        self.fascicles = param['fascicles']
        self.N_electrode = param['N_electrode']
        self.electrodes = param['electrodes']


    ######################
    ## custom the model ##
    ######################

    def reshape_outerBox(self, Outer_D, res="default"):
        """
        Reshape the size of the FEM simulation outer box

        Parameters
        ----------
        outer_D : float
            FEM simulation outer box diameter, in mm, WARNING, this is the only parameter in mm !
        """
        if not self.mesh_file_status:
            self.mesh.reshape_outerBox(Outer_D = Outer_D, res = res)
            self.__update_parameters()

    def reshape_nerve(self, Nerve_D, Length, y_c=0, z_c=0, res="default"):
        """
        Reshape the nerve of the FEM simulation

        Parameters
        ----------
        Nerve_D                 : float
            Nerve diameter, in um
        Length                  : float
            Nerve length, in um
        y_c                     : float
            Nerve center y-coordinate in um, 0 by default
        z_c                     : float
            Nerve z-coordinate center in um, 0 by default
        Perineurium_thickness   :float
            Thickness of the Perineurium sheet surounding the fascicles in um, 5 by default
        """
        if not self.mesh_file_status:
            self.L = Length
            self.Nerve_D
            self.mesh.reshape_nerve(Nerve_D=Nerve_D, Length=Length, y_c=y_c, z_c=z_c, res=res)
            self.__update_parameters()

    def reshape_fascicle(self, Fascicle_D, y_c=0, z_c=0, ID=None, Perineurium_thickness=5, res="default"):
        """
        Reshape a fascicle of the FEM simulation

        Parameters
        ----------
        Fascicle_D  : float
            Fascicle diameter, in um
        y_c         : float
            Fascicle center y-coodinate in um, 0 by default
        z_c         : float
            Fascicle center y-coodinate in um, 0 by default
        ID          : int
            If the simulation contains more than one fascicles, ID number of the fascicle to reshape as in FENICS
        """
        if not self.mesh_file_status:
            self.mesh.reshape_fascicle(Fascicle_D, y_c, z_c, ID, res)
            self.__update_parameters()
            if ID is None:
                if self.Perineurium_thickness == {}:
                    ID = 0
                else:                                   ## To check when not all ID are fixed
                    ID = max(self.Perineurium_thickness) + 1
            self.Perineurium_thickness[ID] = Perineurium_thickness
        
    def add_electrode(self, elec_type, ID=None, res="default", **kwargs):
        """
        
        """
        if not self.mesh_file_status:
            self.mesh.add_electrode(elec_type=elec_type, ID=ID, res=res, **kwargs)


    ########################
    ## setup simulation ##
    ########################

    def setup_simulations(self, comm=None, rank=None):
        """
        
        """
        if self.comm is not None:
            self.comm = comm        
        if self.rank is not None:
            self.rank = rank

        if self.is_meshed and not self.is_sim_ready:
            t0 = time.time()
            # SETTING DOMAINS
            #del self.mesh
            #sim = FEMSimulation(mesh_file=self.mesh_file, elem=self.elem, comm=self.comm, rank=self.rank)
            self.sim = FEMSimulation(mesh_file=self.mesh_file,mesh=self.mesh, elem=self.elem)
            # Outerbox domain
            self.sim.add_domain(mesh_domain=ENT_DOM_offset['Outerbox'],mat_pty=self.Outer_mat)
            # Nerve domain
            self.sim.add_domain(mesh_domain=ENT_DOM_offset['Nerve'],mat_pty=self.Epineurium_mat)
            for i in range(self.N_fascicle):
                self.sim.add_domain(mesh_domain=ENT_DOM_offset['Fascicle']+(2*i),mat_pty=self.Endoneurium_mat)
            for _,(i, elec) in enumerate(self.electrodes.items()):
                if not elec["type"]=="LIFE":
                    self.sim.add_domain(mesh_domain=ENT_DOM_offset['Electrode']+(2*i),mat_pty=self.Electrodes_mat)
            # SETTING INTERNAL BOUNDARY CONDITION (for perineuriums)
            for i in self.fascicles:
                thickness = self.Perineurium_thickness[i]
                f_dom = ENT_DOM_offset['Fascicle']+(2*i)
                self.sim.add_inboundary(mesh_domain=ENT_DOM_offset['Surface'] + f_dom,mat_pty=self.Perineurium_mat, thickness=thickness, in_domains=[f_dom])
            # SETTING BOUNDARY CONDITION
            # Ground (to the external ring of Outerbox)
            self.sim.add_boundary(mesh_domain=1, btype='Dirichlet', value=0)
            # Injected current from active electrode
            # For EIT change in for E in elec_patren:
            for _,(E,active_elec) in enumerate(self.electrodes.items()):
                E_var = 'E'+str(E)
                mesh_domain_3D = self.__find_elec_subdomain(active_elec)
                self.jstims += {self.__find_elec_jstim(active_elec)}
                self.j_electrode[E_var] = 0
                e_dom = ENT_DOM_offset['Surface'] + ENT_DOM_offset['Electrode'] + (2*E)
                self.sim.add_boundary(mesh_domain=e_dom, btype='Neuman', variable=E_var, mesh_domain_3D=mesh_domain_3D)
            self.is_sim_ready = True
        self.preparing_timer += time.time() - t0
            
    
    def set_materials(self, Outer_mat=None, Epineurium_mat=None, Endoneurium_mat=None, Perineurium_mat=None, Electrodes_mat=None):
        """
            Set material files for any domain
        Outer_mat       :str
            Outer box material fname if None not changed, by default None
        Epineurium_mat      :str
            Epineurium material fname if None not changed, by default None
        Endoneurium_mat     :str
            Endoneurium material fname if None not changed, by default None
        Perineurium_mat     :str
            Outer material fname if None not changed, by default None
        Electrodes_mat      :str
            Electrodes material fname if None not changed, by default None
            
        """
        if Outer_mat is not None:
            self.Outer_mat = Outer_mat
        if Epineurium_mat is not None:
            self.Epineurium_mat = Epineurium_mat
        if Endoneurium_mat is not None:
            self.Endoneurium_mat = Endoneurium_mat
        if Perineurium_mat is not None:
            self.Perineurium_mat = Perineurium_mat
        if Electrodes_mat is not None:
            self.Electrodes_mat = Electrodes_mat

    def __find_elec_subdomain(self, elec) -> int:
        """
            Internal use only: 
        """
        if elec["type"]=="LIFE":
            y_e, z_e = elec["kwargs"]['y_c'], elec["kwargs"]['z_c']
            for i in self.fascicles:
                fascicle = self.fascicles[i]
                d_f, y_f, z_f = fascicle["D"], fascicle["y_c"], fascicle["z_c"]
                if (y_e - y_f)**2 + (z_e - z_f)**2 < d_f**2:
                    return 10 + i
        else:
            return 0

    def __find_elec_jstim(self, elec, I=None) -> float:
        """
            Internal use only: return electrical current density in electrode 
            from current valeu and electrode geometry
        """ 
        # Unitary stimulation
        if I is not None:
            self.Istim = I

        if elec["type"] == "CUFF":
            d_e = self.Nerve_D + 2*elec["kwargs"]['contact_thickness']
            l_e = elec["kwargs"]['contact_length']

        elif elec["type"] == "LIFE":
            d_e = elec["kwargs"]['D']
            l_e = elec["kwargs"]['length']
        
        S = pi*(d_e)*(l_e)
        jstim = self.Istim / S
        return jstim

    ###################
    ## Use the model ##
    ###################

    def get_meshes(self):
        """
        Get the different meshes implemented in the model

        Returns
        -------
        list
            list of meshes implemented in the FENICS model file
        """
        if self.is_meshed:
            self.mesh.visualize()

    def build_and_mesh(self):
        """
        Build the geometry and perform meshing process
        """
        if not self.mesh_file_status:
            t0 = time.time()
            self.__update_parameters()
            if self.N_fascicle == 0:
                self.reshape_fascicle(Fascicle_D=self.default_fascicle['D'], y_c=self.default_fascicle['y_c'], z_c=self.default_fascicle['z_c'], res=self.default_fascicle['res'])
            if self.N_electrode == 0:
                self.add_electrode(elec_type=self.default_electrode['elec_type'], x_c=self.default_electrode['x_c'], y_c=self.default_electrode['y_c'], \
                    z_c=self.default_electrode['z_c'], length=self.default_electrode['length'], D=self.default_electrode['D'], res=self.default_electrode['res'])
            self.__update_parameters()
            self.mesh.compute_mesh()
            self.is_meshed = True
            self.mesh.get_info(verbose=True)
            self.meshing_timer += time.time() - t0

    def solve(self, comm=None, rank=None):
        """
        Solve the model
        """
        if not self.is_sim_ready:
            self.setup_simulations()
        
        t0 = time.time()
        # For EIT change in for E in elec_patren:
        for E in range(self.N_electrode):
            for i_elec in self.j_electrode:
                if i_elec == 'E' + str(E):
                    self.j_electrode[i_elec] = self.jstims[E]
                else:
                    self.j_electrode[i_elec] = 0
            self.sim.prepare_sim(**self.j_electrode)
            self.sim_res += [self.sim.solve()]
            self.sim_res[-1].vout.label = 'E'+str(E)
        self.is_computed = True
        self.solving_timer += time.time() - t0

    ######################
    ## results handling ##
    ######################

    def get_potentials(self, x, y, z, E=-1):
        """
        Get the potential on a line to get extracellular potential for axons stimulation.

        Parameters
        ----------
        x   : np.array
            array of x coordinates in the model
        y   : float
            y-coordinate of the axon
        z   : float
            z-coordinate of the axon
        E   : int
            ID of the electrode from witch the potentials should be evaluated

        Returns:
        --------
        array
            All potential for all paramtric sweeps (all electrodes in NRV2 models)\
            (line: electrode selection, column: potential)
        """
        if self.is_computed == True:
            t0 = time.time()
            line = [[x_, y, z] for x_ in x]
            if self.N_electrode==1 or (E < self.N_electrode and E >= 0):
                potentials =  self.sim_res[E].eval(line)*V
            else:
                potentials = []
                for E in range(self.N_electrode):
                    potentials += [self.sim_res[E].eval(line)]
                potentials = np.transpose(np.array(potentials)*V)
            self.access_res_timer += time.time() - t0
            return potentials



""" NeuRon Virtualizer, large scale modeling of Peripheral Nervous System with random stimulation waveforms"""

# Meta information
__title__           = 'NRV'
__project__         = 'NeuRon Virtualizer (NRV)'
__version__         = '0.9.9.alpha'
__date__            = '2023–07–23'
__author__          = 'Florian Kolbl'
__contributors__    = 'Florian Kolbl, Roland Giraud, Louis Regnacq, Thomas Couppey'
__copyright__       = '2023, Florian Kolbl'
__license__         = 'CeCILL'

#################################
# check environnement variables #
#################################
import os
import inspect
import platform
# create a dummy object to locate frameworks path
class DummyClass: pass
nrv_path = os.path.dirname(os.path.abspath(inspect.getsourcefile(DummyClass)))
root_path = nrv_path.replace('/nrv/','')
# check the PATH with os.environ['PATH'], modify bash/zsh profile
current_PATH = os.environ['PATH']
conf_file = '.bashrc'
if platform.system() == 'Darwin':
    conf_file = '.zshrc'
if not (nrv_path+':' in current_PATH or ':'+nrv_path in current_PATH):
    with open(os.path.expanduser('~/'+conf_file), "a") as outfile:
        outfile.write('\n\n\n# >>>>> NRV setup >>>>>\n')
        outfile.write('export PATH="'+nrv_path+':$PATH"\n')
        outfile.write('# <<<<< NRV setup <<<<<\n')
    outfile.close()
    print(conf_file + ' file modified, please source or restart console to be able to used nrv2calm')
# create the environnement variable NRVPATH if it does not exist
if not 'NRVPATH' in os.environ:
    os.environ['NRVPATH'] = nrv_path
# change the permissions on nrv2calm
if not os.access(nrv_path+'/nrv2calm', os.X_OK):
    mode = os.stat(nrv_path+'/nrv2calm').st_mode
    mode |= (mode & 0o444) >> 2    # copy R bits to X
    os.chmod(nrv_path+'/nrv2calm', mode)

# Public interface
from .backend import compileMods
from .backend.parameters import *

from .fmod.materials import *
from .fmod.electrodes import *
from .fmod.stimulus import *
from .fmod.extracellular import *
from .fmod.recording import *
from .fmod.FEM.FEM import *
from .fmod.FEM.COMSOL_model import *
from .fmod.FEM.FENICS_model import *
######### May not be requiered at the end ###############
from .fmod.FEM.mesh_creator.MshCreator import *
from .fmod.FEM.mesh_creator.NerveMshCreator import *
from .fmod.FEM.mesh_creator.NRV_Msh import *
from .fmod.FEM.fenics_utils.FEMSimulation import *    
from .fmod.FEM.fenics_utils.SimParameters import *
from .fmod.FEM.fenics_utils.SimResult import *
########################################################

from .nmod.axons import *
from .nmod.unmyelinated import *
from .nmod.myelinated import *
from .nmod.thin_myelinated import *
from .nmod.fascicles import *
from .nmod.fascicle_generator import *
from .nmod.nerve import *

from .utils.saving_handler import *
from .utils.nrv_function import *
from .utils.cell.CL_postprocessing import *
from .utils.cell.CL_simulations import *
from .utils.cell.CL_discretization import *
from .utils.fascicle.FL_postprocessing import * 

from .optim.CostFunctions import *
from .optim.Optimizers import *
from .optim.Problems import *



nrv.rasterize(sim_results,'V_mem')
nrv.filter_freq(sim_results,'V_mem',10)

plt.figure()
for node in range(len(sim_results['x_rec'])):
	plt.plot(sim_results['t'],sim_results['V_mem_filtered'][node]+100*node,color='k')
plt.xlabel('time (ms)')
plt.yticks([])
plt.ylabel('membrane voltage at Nodes of Ranvier allong axon')
title = 'Axon '+str(k)+', myelination is '+str(sim_results['myelinated'])+', '+str(sim_results['diameter'])+' um diameter'
plt.title(title)
plt.xlim(0,sim_results['tstop'])
plt.tight_layout()
fig_name = folder_name + '/Vmem_axon_'+str(k)+'.pdf'
plt.savefig(fig_name)
plt.close()

nrv.remove_key(sim_results,'V_mem', verbose=verbose)
nrv.remove_key(sim_results,'V_mem_filtered', verbose=verbose)



import numpy as np

from mpi4py import MPI 
import scipy

from dolfinx.fem import (FunctionSpace, Function, Expression)
from dolfinx.io.gmshio import read_from_msh, model_to_mesh
from dolfinx.io.utils import XDMFFile
from dolfinx.geometry import (BoundingBoxTree, compute_collisions, compute_colliding_cells)
from dolfinx.mesh import create_cell_partitioner, GhostMode
import gmsh

from ..mesh_creator.MshCreator import *


from ....backend.MCore import *
from ....backend.file_handler import json_load, json_dump, rmv_ext
from ....backend.log_interface import rise_error, rise_warning, pass_info
from ....backend.NRV_Class import NRV_class

###############
## Functions ##
###############
def is_sim_res(result):
    """
    check if an object is a SimResult, return True if yes, else False

    Parameters
    ----------
    result : object
        object to test

    Returns
    -------
    bool
        True it the type is a SimResult object
    """
    return isinstance(result, SimResult)

def save_sim_res_list(sim_res_list, fname, dt=1):
    """
    save a list of SimResults in a xdmf file
    """
    fname = rmv_ext(fname) + ".xdmf"
    N_list = len(sim_res_list)
    xdmf = XDMFFile(sim_res_list[0].domain.comm, fname, "w")
    xdmf.write_mesh(sim_res_list[0].domain)
    for E in range(N_list):
        xdmf.write_function(sim_res_list[E].vout, E*dt)

def read_gmsh(mesh, comm=MPI.COMM_WORLD, rank=0, gdim=3):
    """
    Given a mesh_file or a MeshCreator returns mesh cell_tags and facet_tags
    (see dolfinx.io.gmshio.model_to_mesh for more details)
    NB: copy of dolfinx.io.gmshio.read_from_msh verbose from gmsh
    
    Parameters
    ----------
    mesh_file : str, 
        Path and name of mesh file 
        NB: extention is not required but must be a .msh file
    comm            : tupple (str, int)
        The MPI communicator to use for mesh creation
    rank            : tupple (str, int)
        The rank the Gmsh model is initialized on

    Returns
    -------
    output  :  tuple(3)
        (domain, cell_tag, facet_tag)

    """
    if comm.rank == rank:
        if isinstance(mesh, str):
            mesh_file = rmv_ext(mesh) + ".msh"
            clear_gmsh()
            gmsh.initialize()
            gmsh.option.setNumber("General.Verbosity", 2)
            gmsh.model.add("Mesh from file")
            gmsh.merge(mesh_file)
            output = model_to_mesh(gmsh.model, comm=comm, rank=rank, gdim=gdim)
            gmsh.finalize()
        elif is_MshCreator(mesh):
            output = model_to_mesh(mesh.model, comm=comm, rank=rank, gdim=gdim)
        else: 
            rise_error('mesh should be either a filename or a MeshCreator')
    return output


def domain_from_meshfile(mesh_file):
    """
        return only the domain from mesh_file
    
    Parameters
    ----------
    mesh_file : str
        Path and name of mesh file 
        NB: extention is not required but must be a .msh file

    Returns
    -------
        output  :
            domain of the mesh contain in the mesh_file
    """
    return read_gmsh(mesh_file)[0]



def V_from_meshfile(mesh_file, elem=('Lagrange', 1)):
    mesh = domain_from_meshfile(mesh_file)
    V = FunctionSpace(mesh, elem)
    return V
    

class SimResult(NRV_class):
    """
    Result of a FEMSimulation. 
    Store the resulting function space giving the possibility to apply basic mathematical
    operations on multiple results
    """
    def __init__(self, mesh_file="", domain=None, elem=('Lagrange', 1), V=None, vout=None,comm=MPI.COMM_WORLD):
        """
        initialisation of the SimParameters:
        
        Parameters
        ----------
        mesh_file       :str
            mesh directory and file name: by default ""
        domain       : None or mesh
            mesh domain on which the result is defined, by default None
        elem            :tupple (str, int)
            if None, ('Lagrange', 1), else (element type, element order), by default None
        V       : None or dolfinx.fem.FunctionSpace
            FunctionSpace on which the result is defined, by default None
        vout       : None or dolfinx.fem.Function
            Function resulting from the FEMSimulation, by default None
        comm            :int
            The MPI communicator to use for mesh creation, by default MPI.COMM_WORLD
        """    
        super().__init__()
        self.type = "simresult"
        self.mesh_file = mesh_file

        self.domain = domain
        self.V = V
        self.elem = elem
        self.vout = vout
        self.comm = comm

    def set_sim_result(self, mesh_file="", domain=None, V=None, elem=None,vout=None, comm=None):
        if mesh_file != "":
            self.mesh_file = mesh_file
        if domain is not None:
            self.domain = domain
        if elem is not None:
            self.elem = elem
        if V is not None:
            self.V = V
        if vout is not None:
            self.vout = vout
        self.comm = comm     
    
    def save(self, file, ftype='xdmf', overwrite=True):
        if ftype == 'xdmf':
            fname = rmv_ext(file) + '.xdmf'
            with XDMFFile(self.comm, fname, "w") as file:
                if not overwrite:
                        file.parameters.update(
                            {
                            "functions_share_mesh": True,
                            "rewrite_function_mesh": False
                            })
                else:
                    file.write_mesh(self.domain)
                file.write_function(self.vout)
        else:
            fname = rmv_ext(file) + '.sres'
            mdict = {"mesh_file":self.mesh_file, "element": self.elem,"vout":self.vout.vector[:]}
            scipy.io.savemat(fname, mdict)
            return mdict

    def load(self, file):
        fname = rmv_ext(file) + '.sres'
        mdict = scipy.io.loadmat(fname)
        self.mesh_file = mdict['mesh_file'][0]
        self.elem = (mdict['element'][0].strip(), int(mdict['element'][1]))
        if self.domain is None:
            self.domain = domain_from_meshfile(self.mesh_file)
            self.V = FunctionSpace(self.domain, self.elem)
        self.vout = Function(self.V)
        self.vout.vector[:] = mdict['vout']

    def save_sim_result(self, file, ftype='xdmf', overwrite=True):
        rise_warning('save_sim_result is a deprecated method use save')
        self.save(file=file,ftype=ftype,overwrite=overwrite)
    def load_sim_result(self, data='sim_result.json'):
        rise_warning('load_sim_result is a deprecated method use load')
        self.load(data=data)

    #############
    ## methods ##
    #############

    def vector(self):
        return self.vout.x.array

    def aline_V(self, res2):
        """

        """
        if is_sim_res(res2):
            if self.mesh_file == res2.mesh_file:
                expr = Expression(self.vout, res2.V.element.interpolation_points())
                self.vout = Function(res2.V)
                self.vout.interpolate(expr)
                self.V = res2.V
            else:
                rise_error("To aline mesh function reslults must have the same meshfile")
        else:
            rise_error("Mesh function alinment must be done with SimResult")


    def eval(self, X):
        """
        Eval the result field at X position
        """
        N = len(X)
        to_round = False
        tree = BoundingBoxTree(self.domain, self.domain.geometry.dim)
        cells_candidates = compute_collisions(tree, X)
        cells_colliding = compute_colliding_cells(self.domain, cells_candidates, X)
        cells = []
        for i in range(N):
            cell = cells_colliding.links(i)
            if len(cell)==0:
                if i == N-1 and i>0:
                    to_round = True
                else:
                    rise_warning(X[i], " not found in mesh, value of ", X[i-1], " reused")
                cells += [cells[-1]]
            else:
                cells += [cell[0]]

        values = self.vout.eval(X, cells)
        if N>1:
            values = values[:,0]
        if to_round:
            X_1 = X[-1]
            X_1[0] = round(X_1[0], 1)
            value_1 = self.eval([X_1])
            values[-1] = value_1[0]
        return values

    

    #####################
    ## special methods ##
    #####################

    """
    def __abs__(self):
        res = SimResult(mesh_file=self.mesh_file, V=self.V, domain=self.domain, elem=self.elem, comm=self.comm)
        res.vout = df.project(abs(self.vout), self.V, solver_type=self.solver,\
            preconditioner_type=self.preconditioner, form_compiler_parameters=self.compiler_parameters)
        return res
    """
    def __neg__(self):
        expr = Expression(-self.vout, self.V.element.interpolation_points())
        res = SimResult(mesh_file=self.mesh_file, V=self.V, domain=self.domain, elem=self.elem, comm=self.comm)
        res.vout = Function(self.V)
        self.vout.interpolate(expr)
        return res

    def __add__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            expr = Expression(self.vout + b.vout, self.V.element.interpolation_points())
        else:
            expr = Expression(self.vout + b, self.V.element.interpolation_points())

        C = SimResult(mesh_file=self.mesh_file, V=self.V, domain=self.domain, elem=self.elem, comm=self.comm)
        C.vout = Function(self.V)
        C.vout.interpolate(expr)
        return C

    def __sub__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            expr = Expression(self.vout - b.vout, self.V.element.interpolation_points())
        else:
            expr = Expression(self.vout - b, self.V.element.interpolation_points())

        C = SimResult(mesh_file=self.mesh_file, V=self.V, domain=self.domain, elem=self.elem, comm=self.comm)
        C.vout = Function(self.V)
        C.vout.interpolate(expr)
        return C

    def __mul__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            expr = Expression(self.vout * b.vout, self.V.element.interpolation_points())
        else:
            expr = Expression(self.vout * b, self.V.element.interpolation_points())

        C = SimResult(mesh_file=self.mesh_file, V=self.V, domain=self.domain, elem=self.elem, comm=self.comm)
        C.vout = Function(self.V)
        C.vout.interpolate(expr)
        return C

    def __radd__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            expr = Expression(b.vout + self.vout, self.V.element.interpolation_points())
        else:
            expr = Expression(b + self.vout, self.V.element.interpolation_points())

        C = SimResult(mesh_file=self.mesh_file, V=self.V, domain=self.domain, elem=self.elem, comm=self.comm)
        C.vout = Function(self.V)
        C.vout.interpolate(expr)
        return C

    def __rsub__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            expr = Expression(b.vout - self.vout, self.V.element.interpolation_points())
        else:
            expr = Expression(b - self.vout, self.V.element.interpolation_points())

        C = SimResult(mesh_file=self.mesh_file, V=self.V, domain=self.domain, elem=self.elem, comm=self.comm)
        C.vout = Function(self.V)
        C.vout.interpolate(expr)
        return C

    def __rmul__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            expr = Expression(b.vout * self.vout, self.V.element.interpolation_points())
        else:
            expr = Expression(b * self.vout, self.V.element.interpolation_points())

        C = SimResult(mesh_file=self.mesh_file, V=self.V, domain=self.domain, elem=self.elem, comm=self.comm)
        C.vout = Function(self.V)
        C.vout.interpolate(expr)
        return C

    def __eq__(self, b):
        if is_sim_res(b):
            if b.mesh_file == self.mesh_file:
                if np.allclose(self.vector(), b.vector()):
                    return True
        return False

    def __ne__(self, b): # self != b
        return not self == b



from ..backend.log_interface import rise_error, rise_warning, pass_info
from ..backend.file_handler import json_dump
from ..backend.NRV_Class import NRV_class, load_any

class CostFunction(NRV_class):
    """
    A class to define cost from position input vector

    Parameters
    ----------
    context_modifier   : funct
        function creating a context from a particle position, parameters:
            position        : particle position in each dimensions (array)
            t_sim           : simulation time (float)
            dt              : time step of the context (float)
            kwargs          : keys word arguments
        returns
            context        : values of a context (array)
    simulate_context   : funct
        function which perform the NEURON simulation, parameters:
            context        : context to use for the simulation (array)
            t_sim           : simulation time (float)
            dt              : time step of the context (float)
            kwargs          : keys word arguments
        returns
            results        : axon simulaiton results (dictionary) 
    residual            : funct
        function calculating a residual from a axon simulation results, parameters:
            results        : axon simulaiton results (dictionary)
            kwargs          : keys word arguments
        returns
            cost              : residual value (float)
    kwargs_gw           : dict
        key word arguments of context_modifier function, by default {}
    kwargs_sw           : dict
        key word arguments of simulate_context function, by default {}
    kwargs_r            : dict
        key word arguments of residual function, by default {}
    t_sim           : float
        time of the simulation on wich the residual will be calculated in ms (ms), by default 100ms
    dt              : float
        simulation time stem for neuron (ms), by default 1us
    part_filter         : funct or None
        function wich return a filtered postition from a position, if None, position is unfiltered,
        by default None
    saver         : funct or None
        function added at the end after calculating the cost for personalized savings, if None,
        nothing will be saved,  by default None

    """
    def __init__(self, static_context, context_modifier, residual, kwargs_gw={}, simulate_context = None,
        kwargs_sw={}, kwargs_r={}, t_sim=100, dt=0.005, filter=None, saver=None, 
        file_name='cost_saver.csv'):
        self.static_context = static_context
        self.context_modifier = context_modifier
        #self.simulate_context = simulate_context
        self.residual = residual
        self.kwargs_gw = kwargs_gw
        self.kwargs_sw = kwargs_sw
        self.kwargs_r = kwargs_r
        self.t_sim = t_sim
        self.dt = dt
        self.filter = filter
        self.saver = saver
        self.file_name = file_name

    def simulate_context(self, context):
        print('pioup')
        results = context.simulate(t_sim = self.t_sim, loaded_footprints=True)
        return results

    def __call__(self, X):
        # Filter
        if self.filter is not None:
            X_ = self.filter(X)
        else:
            X_ = X

        # Interpolation
        simulation_context = self.context_modifier(X_, self.static_context)

        # Simulation
        results = self.simulate_context(simulation_context)

        # Cost calculation
        cost = self.residual(results, **self.kwargs_r)

        # Savings
        if self.saver is not None:
            data = {'position':X, 'context':simulation_context, 'results':results, 'cost':cost}
            self.saver(data, file_name=self.file_name)
        return cost




"""
NRV-Nerves
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import numpy as np
import matplotlib.pyplot as plt
from ..backend.log_interface import rise_error, rise_warning, pass_info

from .fascicles import *
from ..backend.NRV_Class import NRV_class

#################
## Nerve class ##
#################
class nerve(NRV_class):
    """A nerve in NRV2 is defined as:
        - a list of fascicles
        - a list of materials
        - a kind of extracellular context (analytical or FEM based)

    a nerve can be instrumented by adding couples of electrodes+stimulus
    """
    def __init__(self, Length=10000, Outer_D=5,  Nerve_D=4000, y_c=0, z_c=0,\
            dt=0.001, mesh_shape='plateau_sigmoid', alpha_max=0.3, d_lambda=0.1, T=None, ID=0, threshold=-40,\
            Adelta_limit=1,**kwargs):
        """
        Instanciates an empty nerve.

        Parameters
        ----------
        dt              : float
            simulation time stem for Neuron (ms), by default 1us
        Nseg_per_sec    : float
            number of segment per section in Neuron. If set to 0, the number of segment per section is calculated with the d-lambda rule
        freq            : float
            frequency of the d-lambda rule (Hz), by default 100Hz
        freq_min        : float
            minimum frequency for the d-lambda rule when meshing is irregular, 0 for regular meshing
        v_init          : float
            initial value for the membrane voltage (mV), specify None for automatic model choice of v_init
        T               : float
            temperature (C), specify None for automatic model choice of temperature
        ID              : int
            axon ID, by default set to 0
        threshold       : int
            membrane voltage threshold for spike detection (mV), by default -40mV
        Adelta_limit    : float
            limit diameter between A-delta models (thin myelinated) and myelinated models for axons
        """
        super().__init__()
        self.type = "nerve"
        self.L = Length
        self.y_c = y_c
        self.z_c = z_c

        self.Outer_D = (Outer_D*mm)

        self.L = L
        self.FEM = FEM
        ### empty attributes for later use
        self.N_ax = 0
        self.fascicles_IDs = []
        self.fascicles = []
        self.electrodes_IDs = []
        self.electrodes = []
        self.stimuli_IDs = []
        self.stimuli = []
        self.materials = {}
        self.extracellular_context = None

    ## save/load methods
    def save(self, fname, extracel_context=False, intracel_context=False, rec_context=False):
        """
        Save a nerve in a json file

        Parameters
        ----------
        fname : str
            name of the file to save the nerve
        extracel_context: bool
            if True, add the extracellular context to the saving
        intracel_context: bool
            if True, add the intracellular context to the saving
        """
        if MCH.do_master_only_work():
            # copy everything into a dictionnary
            nerve_config = {}
            nerve_config['ID'] = self.ID

            if intracel_context:
                nerve_config['L'] = self.L

            if extracel_context:
                nerve_config['L'] = self.L

            if rec_context:
                nerve_config['record'] = self.record

            # save the dictionnary as a json file
            json_dump(nerve_config, fname)

    def load(self, fname, extracel_context=False, intracel_context=False, rec_context=False):
        """
        Load a nerve configuration from a json file

        Parameters
        ----------
        fname           : str
            path to the json file describing a nerve
        extracel_context: bool
            if True, load the extracellular context as well
        intracel_context: bool
            if True, load the intracellular context as well
        """
        if type(fname) == str:
            results = json_load(fname)
        else: 
            results = fname
        self.ID = results['ID']

        if intracel_context:
            self.L = results['L']
        if extracel_context:
            self.L = results['L']
            
        if rec_context:
            self.record = results['record']

    def add_fascicle(self, fascicle):
        """
        Add a fascicle to the list of fascicles

        Parameters
        ----------
        fascicle : object
            fascicle to add to the nerve struture
        """
        if not is_fascicle(fascicle):
            rise_warning('Only fascile object can be added with add fascicle method: nothing added')
        else:
            if fascicle.ID in self.fascicles_IDs:
                pass_info('Fascicle ' +str(fascicle.ID)+' already in the nerve: will be replace')
                i_fasc = np.index(fascicle.ID)
                self.fascicles[i_fasc]
            else:
                self.fascicles_IDs += [fascicle.ID]
                i_fasc = len(self.fascicles_IDs) + 1
                self.fascicles += [fascicle]
            self.fascicles[i_fasc].L = self.L
            
        
    def __build_extracellular_context(self):
        """
        Private method to create the correct extracellular context
        """
        extracellular_context_list = []
        for fasc in self.fascicles:
            if fasc.extra_stim is not None:
                extracellular_context_list += fasc.extra_stim
                fasc.extra_stim = None

    def simulate(self, *args):
        """
        Simulate the nerve with the proposed extracellular context. Top level method for large scale neural simulation.

        Parameters
        ----------


        Warning
        -------
        calling this method can result in long processing time, even with large computational ressources.
        Keep aware of what is really implemented, ensure configuration and simulation protocol is correctly designed.
        """
        raise NotImplementedError
        # step 1: create FEM geometry, mesh and compute
        self.__build_extracellular_context()
        # step 2: simulate all fascicles
        for fascicle in self.fascicles:
            fascicle.simulate()



from cmath import phase

from .NerveMshCreator import *
from ...electrodes import *
from ....nmod.fascicles import *
from ....backend.log_interface import rise_error, rise_warning, pass_info
from ....utils.units import *

def mesh_from_electrode(elec, mesh=None, Length=10000, Outer_D=5,  Nerve_D=4000, y_c=0, z_c=0, res="default"):
    """
    returns the corresponding mesh from a nrv.facsicle
    """
    if not is_FEM_electrode(elec):
        rise_warning('Only FEM electrode can be added to a mesh')
    else:
        if mesh is None:
            mesh = NerveMshCreator(Length=Length, Outer_D=Outer_D, Nerve_D=Nerve_D, y_c=y_c, z_c=z_c)
        elec.parameter_model(mesh, res=res)
    return mesh

def mesh_from_fascicle(fascicle, mesh=None, Length=10000, Outer_D=5,  Nerve_D=4000, y_c=0, z_c=0,\
    res_nerve="default", res_fasc="default", res_ax="default", res_elec="default"):
    """
    returns the corresponding mesh from a nrv.facsicle
    """
    if mesh is None:
        mesh = NerveMshCreator(Length=Length, Outer_D=Outer_D, Nerve_D=Nerve_D, y_c=y_c, z_c=z_c)
        if res_nerve != "default":
            mesh.reshape_nerve(res=res_nerve)
    
    mesh.reshape_fascicle(D=fascicle.D, y_c=fascicle.y_grav_center, z_c=fascicle.z_grav_center,\
        res=res_fasc)
    
    for i_ax in range(fascicle.N):
        ax_d=round(fascicle.axons_diameter[i_ax], 3)
        ax_y_c=round(fascicle.axons_y[i_ax], 3)
        ax_z_c=round(fascicle.axons_z[i_ax], 3)
        if isinstance(res_ax, str) and res_ax != "default":
            res_ax = eval(str(fascicle.axons_diameter[i_ax]) + res_ax)
        mesh.reshape_axon(D=ax_d, y_c=ax_y_c, z_c=ax_z_c, res=res_ax)
    
    if fascicle.extra_stim is not None:
        N_contact = 0 
        if np.iterable(res_elec) and not isinstance(res_elec, str):
            if len(res_elec) == len(fascicle.extra_stim.electrodes):
                res = res_elec
            else:
                rise_warning("length of res_elec and fascicle.extra_stim.electrodes does not match",\
                    "only first value kept")
                res = [res_elec[0] for k in range(len(fascicle.extra_stim.electrodes))]
        else:
            res = [res_elec for k in range(len(fascicle.extra_stim.electrodes))]
        for i_elec, elec in enumerate(fascicle.extra_stim.electrodes):
            if (not elec.is_multipolar) or N_contact<=0:
                mesh = mesh_from_electrode(elec, mesh=mesh, res=res[i_elec])
                if elec.is_multipolar:
                    N_contact = elec.N_contact
            else:
                N_contact -= 1

                

    return mesh

def mesh_from_nerve(nerve, mesh=None):
    rise_warning('not implemented yet')



"""
log_interface handeling
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import os
import logging
import configparser
import sys
from icecream import ic
from .MCore import *

from .parameters import parameters




class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

dir_path = os.environ['NRVPATH'] + '/_misc'
# logging config
logging.basicConfig(filename=dir_path+'/log/NRV.log', level=logging.INFO, format=\
    '%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')


def rise_error(*args, out=1, **kwargs):
    """
    Rises and error to the log and to the prompt for the master process (process ID 0) in case of
    parallel computing. This function exit the programm.

    Parameters
    ----------
    *args   :
        anything to pass as error message
    out     : int
        programm exit value, must be strictly positive. By default equal to 1
    """
    verbose = True
    if 'verbose' in kwargs:
        verbose = kwargs['verbose']

    message = ''
    for arg in args:
        message += str(arg)
    if MCH.is_alone():
        if parameters.LOG_Status:
            logging.error(message)
        if verbose and parameters.VERBOSITY_LEVEL>=1:
            print(bcolors.FAIL + 'NRV ERROR: '+ message+ bcolors.ENDC)
    else:
        if parameters.LOG_Status:
            logging.error('NRV ERROR: '+ message + '\n encountered in process '+str(MCH.rank)+' out of '+str(MCH.size))
    if out == 0:
        out = 1
    sys.exit(out)


def rise_warning(*args, abort=False, **kwargs):
    """
    Rises a warning to the log and to the prompt for the master process (process ID 0) in case of
    parallel computing. This function can exit the programm.

    Parameters
    ----------
    *args   :
        anything to pass as warning message
    abort   : boolean
        if true, the programm exits with the value O (no error) and further computation are avoided,
        by default set to False.
    """
    verbose = True
    if 'verbose' in kwargs:
        verbose = kwargs['verbose']
    message = ''
    for arg in args:
        message += str(arg)
    if MCH.is_alone():
        if parameters.LOG_Status:
            logging.warning('NRV WARNING: '+ message)
        if verbose and parameters.VERBOSITY_LEVEL>=2:
            print(bcolors.WARNING + 'NRV WARNING: '+ message+ bcolors.ENDC)
    else:
        if parameters.LOG_Status:
            logging.warning('NRV WARNING: '+ message + '\n encountered in process '+str(MCH.rank)+' out of '+str(MCH.size))
        if MCH.do_master_only_work() and parameters.VERBOSITY_LEVEL>=2:
            print(bcolors.WARNING + 'NRV WARNING: '+ message+ bcolors.ENDC)
            sys.stdout.flush()
    if abort:
        sys.exit(0)

def pass_info(*args, **kwargs):
    """
    Pass an info to the log and to the prompt for the master process (process ID 0) in case of
    parallel computing.

    Parameters
    ----------
    *args   :
        anything to pass as info
    """

    verbose = True
    if 'verbose' in kwargs:
        verbose = kwargs['verbose']
    message = ''
    for arg in args:
        message += str(arg)
    if MCH.is_alone():
        if parameters.LOG_Status:
            logging.info('NRV INFO: '+ message)
        if verbose and parameters.VERBOSITY_LEVEL>=3:
            print('NRV INFO: '+ message)
    else:
        if parameters.LOG_Status:
            logging.info('NRV INFO: '+ message + '\n from process '+str(MCH.rank)+' out of '+str(MCH.size))
    

def pass_debug_info(*args, **kwargs):
    """
    Pass an info to the log and to the prompt for the master process (process ID 0) in case of
    parallel computing.

    Parameters
    ----------
    *args   :
        anything to pass as info
    """

    verbose = True
    if 'verbose' in kwargs:
        verbose = kwargs['verbose']
    message = ''
    for arg in args:
        message += str(arg)
    if MCH.is_alone(): 
        if parameters.LOG_Status and parameters.VERBOSITY_LEVEL>=4: 
            logging.info('NRV DEBUG: '+ message)
        if verbose and parameters.VERBOSITY_LEVEL>=4:
            print('NRV DEBUG: '+ message)
    else:
        if parameters.LOG_Status:
            logging.info('NRV DEBUG: '+ message + '\n from process '+str(MCH.rank)+' out of '+str(MCH.size))

def progression_popup(current, max_iter, begin_message='', end_message='', endl=''):
    """
    Displays the progression on prompt, for single process only, nothing saved to the log

    Parameters
    ----------
    current         : int
        current iteration
    max_iter        : int
        maximum iteration number
    begin_message   : str
        additional message placed at the start of the progression line
    end_message     : str
        additional message placed after iteration and max at the end of the progression line
    endl            : str
        line termination, by default empty string
    """
    if MCH.is_alone():
        print(begin_message + f"{current+1}" + '/' + str(max_iter) + end_message, end=endl)
        if current == max_iter - 1:
            print('\n')

def prompt_debug(*args):
    """
    outputs for debug, simple redirection to icecream without the installation (see icecream help).
    Could be interesting to redirect also to the log...

    Parameters
    ----------
    *args   :
        anything to pass to icecrean
    """
    ic(*args)
